r"""We deal with the 196884-dimensional representation of the monster.

The monster :math:`\mathbb{M}` has a 196884-dimensional 
representation :math:`\rho` with coefficients in 
:math:`\mathbb{Z}[\frac{1}{2}]`, see :cite:`Con85`. 
Representation  :math:`\rho` is called 
:math:`196884_x` in  :cite:`Con85`. We obtain a modular 
representation :math:`\rho_p` of :math:`\mathbb{M}` by 
reducing the matrix coefficients of the representation 
:math:`\rho` modulo an odd number :math:`p`. 
The representation  :math:`\rho_p` can be implemented very
efficiently if :math:`p + 1` is a small power of two.
The current version of the *mmgroup* package supports the 
representations :math:`\rho_p` of :math:`\mathbb{M}` for 
``p = 3, 7, 15, 31, 127, 255``. In the sequel 
we are a bit sloppy, calling :math:`\rho_p` a vector space also
in cases ``p = 15, 255``. The user may select these values for
obtaining representations modulo ``5`` and ``17``, respectively.


One purpose of the ``mmgroup`` package is to give the user the 
capability to compute in the ``196884``-dimensional representation 
:math:`\rho_p` of the monster group :math:`\mathbb{M}`.

For calculating in :math:`\rho_p` the user may create an instance 
of class |MM| that models an element ``g`` of the monster group  
:math:`\mathbb{M}` as described in section :ref:`mmgroup-label`. 
Then the user may create an instance ``v`` of class |MMVector| 
that models a vector in :math:`\rho_p`,  with ``p`` as above. 
The element ``g`` of :math:`\mathbb{M}` acts on the 
vector ``v`` by right multiplication.




Creating basis vectors of the representation space :math:`\rho_p`
.................................................................

By construction of :math:`\rho_p` it is natural to label a basis 
vector of :math:`\rho_p` by a tuple ``(tag, i0, i1)`` with integers 
``i0, i1``, as  described in the next section.  Here ``tag`` is a 
single capital letter and indices ``i0``, ``i1`` are unsigned integers; 
these three quantities describe a basis vector of the space ``V``.

To create a basis vector of  :math:`\rho_p` corresponding to the
tuple ``(tag, i0, i1)`` the user may call the constructor
``MMVector(p, tag, i0, i1)``.

For a fixed characteristic ``p`` the function ``MMV(p)`` returns 
an object correponding to the vector space  :math:`\rho_p`.
Thus the sequence

 .. code-block:: python

  >>> from mmgroup import MMV  
  >>> V = MMV(p)  
  >>> v = V(tag, i0, i1)

creates the same basis vector ``v`` as the 
statement ``v = MMVector(p, tag, i0, i1)``. 




Description of the basis vectors of the representation space
.............................................................


The following table shows the tags available for describing basis
vectors. The column ``Vector`` in that table shows the basis vector  
in the notation of :cite:`Seysen20`. The column ``Tuple`` shows the 
basis vector as a tuple ``(tag, i0, i1)``.

.. _table-vector-tags:

.. table:: Basis vectors of the representation of the monster
  :widths: 10 15 75

  ============================ ============= =================================
  Vector                       Tuple         Remarks
  ============================ ============= =================================
  :math:`(ij)_1`               ``('A',i,j)`` ``0 <= i,j < 24``; we have
                                             :math:`(ij)_1 = (ji)_1`.

  :math:`X_{ij}`               ``('B',i,j)`` ``0 <= i,j < 24, i != j``; we 
                                             have :math:`X_{ij} = X_{ji}`.  

  :math:`X^+_{ij}`             ``('C',i,j)`` ``0 <= i,j < 24, i != j``; we 
                                             have :math:`X^+_{ij} = X^+_{ji}`.

  :math:`X^+_{d \cdot \delta}` ``('T',o,s)`` ``o`` and ``s`` can be obtained 
                                             as follows. Let,  
                                             ``x = SubOctad(d, delta)``. Then
                                             ``x.vector_tuples()`` returns
                                             the tuple ``(1, 'T', o, s)``.

                                             :math:`d \in \mathcal{P}`,
                                             :math:`\delta \in \mathcal{C^*}`,
                                             :math:`d` an octad, 
                                             :math:`\delta \subset d`,
                                             :math:`\delta` even. 

                                             We have ``0 <= o < 759``,
                                             ``0 <= s < 64``. 

  :math:`X^+_{d \cdot i}`      ``('X',d,i)`` ``0 <= d < 2048``, 
                                             ``0 <= i < 24``,

                                             ``d`` represents the element
                                             ``PLoop(d)`` of the Parker
                                             loop :math:`\mathcal{P}`

  :math:`d^+ \otimes_1 i`      ``('Z',d,i)`` ``d`` and ``i``  as in case              
                                             ``('X',d,i)``

  :math:`d^- \otimes_1 i`      ``('Y',d,i)`` ``d`` and ``i``  as in case              
                                             ``('X',d,i)``
  ============================ ============= =================================

**Remarks**

  * The space :math:`\rho_p` is an orthogonal space. 
    Two basis vectors are orthogonal except when equal or opposite.
    The basis vectors :math:`(ij)_1` have norm ``2`` in case
    :math:`i \neq j`. All other basis vectors have norm ``1``. 

  * In the tuple ``('T',o,s)`` the integers ``o`` and ``s`` describe
    the instance ``SubOctad(o, s)`` of class |XLeech2|. Here 
    ``o`` and ``s`` may be anything that is accepted as input for
    ``SubOctad(o, s)``

  * In the tuples ``('X',d,i)``, ``('Y',d,i)``, ``('Z',d,i)``,
    the input ``d`` may be an instance of class |PLoop| or 
    anything such that ``PLoop(d)`` is an instance of class |PLoop|.
    For an instance ``d`` of class |PLoop| we have:

      * ``('X',d,i) == -('X',-d,i) ==  ('X',~d,i)`` ,
      * ``('Y',d,i) == -('Y',-d,i) == -('Y',~d,i)`` ,
      * ``('Z',d,i) == -('Z',-d,i) ==  ('Z',~d,i)`` .

    The value of ``('X',~d,i)`` has been set by convention;
    the other equations are forced.

  * The basis vector :math:`d^+ \otimes_1 i` in :cite:`Seysen20` is 
    equal to the basis vector 
    :math:`(-1)^{|d/4|} \cdot d^+ \otimes_1 i` in :cite:`Con85`. This 
    modification simplifies the formulas for the operation of the 
    element :math:`\xi` of  :math:`\mathbb{M}` on :math:`\rho_p`.

  * A first or second index following a tag may be ``'r'``, indicating 
    a random integral index. Any omitted index after a given index 
    ``'r'`` is interpreted as ``'r'``.  For tags ``'A', 'B', 'C', 'T'`` 
    the second index should  be ``'r'`` or omitted if the first index 
    is ``'r'``, since in these cases the two indices are not 
    sufficiently independent.


Representing a vector as a list of tuples
.........................................

For creating an arbitrary 
(but yet sparse) vector in :math:`\rho_p` the user may call
``MMVector(p, data_list)``, where ``data_list`` is a list of
tuples ``(factor, tag, i0, i1)``. Here ``(tag, i0, i1)``
describes a unit vector and the (optional) integer ``factor`` is 
a scalar factor for that unit vector. Then the constructor returns 
the linear combination of the unit vectors given in the list.

In order to generate a random multiple of a basis vector, inside a 
list of tuples, the (optional) component ``factor`` of a tuple can 
be set to one of the following values:

      * ``'u'``: equivalent to ``1`` 
      * ``'s'``: a factor ``1`` or ``-1`` selected at random
      * ``'n'``: a random nonzerodivisor modulo ``p`` 
      * ``'r'``: a random integer modulo ``p`` 


Operations on vector in the representation :math:`\rho_p` 
.........................................................

Vector addition, subtraction and scalar multiplication can be done with
operators ``+``, ``-`` and ``*`` as usual. Elements of the monster group
operate on vectors by right multiplication. Only vectors modulo
the same characteristic ``p``  can be added. 

The following code example generates an element ``g`` of the
monster group and a vector ``v`` in the vector space  :math:`\rho_3` 
and displays the result ``g * v`` of the operation of ``g`` on ``v``. 
Note that a basis vector ``(tag, i0, i1)`` is displayed in the 
form ``tag_i0_i1``. For a displayed index ``i0`` or ``i1`` the 
suffix ``h`` means hexadecimal notation.


.. code-block:: python

  >>> # An instance of class MM represents an element of the monster
  >>> from mmgroup import MM
  >>> # Function MMV is used to create a representation space
  >>> from mmgroup import MMV
  >>> # Create representation space V for the monster (modulo 3)
  >>> V = MMV(3)
  >>> # Create an element g of the monster group
  >>> g = MM([('d', 0x123), ('p', 217821225)])
  >>> # Create a vector v in the representation space V 
  >>> v = V([('T', 712, 13), (2, 'X', 0x345, 13)])
  >>> # Let g operate on v by right multiplication
  >>> print(v * g)
  MV<3;-T_444_0fh+X_244h_11>
  

Special tags for creating vectors in the representation :math:`\rho_p` 
.......................................................................

Apart from the standard tags ``A``, ``B``, ``C``, ``T``, ``X``, ``Y``,
and ``Z``, the constructor of class |MMVector| acceps a variety of 
special tags. Details are given in the following list:

.. table:: Special tags for constructing a vector
   :widths: 25 75

   ========================= ============================================== 
   tag, i0, i1               Evaluates to                               
   ========================= ============================================== 
   ``('D', i0)``             Shorthand for ``('A', i0, i0)``   
   ------------------------- ---------------------------------------------- 
   ``('I', i0, i1)``         Shorthand for the sum of the basis vectors 

                             ``('A', i0, i0)`` + ``('A', i1, i1)`` - 
                             ``('A', i0, i1)`` - ``2 ('B', i0, i1)``,

                             see remark below. 
   ------------------------- ---------------------------------------------- 
   ``('J', i0, i1)``         Shorthand for the sum of the basis vectors 

                             ``('A', i0, i0)`` + ``('A', i1, i1)`` - 
                             ``('A', i0, i1)`` + ``2 ('B', i0, i1)``,

                             see remark below. 
   ------------------------- ---------------------------------------------- 
   ``('U')``                 Tag ``'U'`` (without any further parameters)
                             stands for the sum 
                             
                             ``('A', 0, 0)`` + ``('A', 1, 1)`` + ...
                             +  ``('A', 23, 23)``.
   ------------------------- ---------------------------------------------- 
   ``('E', i)``              This is the basis vector with
                             *linear* index ``i``. 

                             See the following subsection for details.              
   ------------------------- ---------------------------------------------- 
   ``('S', data)``           Here ``data`` is an array-like object (as
                             defined in the numpy package) that encodes
                             a vector in sparse representation as an 
                             array of unsigned 32-bit integers. 

                             This is for internal use. For details see
                             the following subsection: 

                             Sparse representation of vectors 
                             in :math:`\rho_p`.            
   ------------------------- ---------------------------------------------- 
   ``('V', data)``           Here ``data`` is an array like object (as
                             defined in the numpy package) that encodes  
                             a one-dimensional array of integers of
                             length 196884. The order of the entries
                             is as in the next section.              
   ------------------------- ---------------------------------------------- 
   ``('R')``                 Tag ``'R'`` (without any further parameters)
                             stands for the generation of a uniform
                             distributed random vector.
   ------------------------- ---------------------------------------------- 
   ``(i, v)``, ``i`` integer Here ``i`` is an integer and ``v`` is a
                             vector, i.e.an instance of class
                             |MMVector|. Then the vector ``i * v``
                             is generated. This is useful for  
                             extending the modulus ``p`` of a 
                             vector, see  remark below.
   ========================= ============================================== 

Remarks

The vectors labelled by ``('I', i0, i1)`` and ``('J', i0, i1)`` are 
axes of the elements :math:`x_\delta` and :math:`x_{-1} x_\delta` of 
the monster, respecively, see  :cite:`Con85`. The centralizers of 
these elements and also of their axes in the monster have structure 
:math:`2 \cdot B`, where :math:`B` is the Baby Monster, see 
:cite:`Asc86`, :cite:`Con85`, :cite:`Iva09` for background. E.g. 
the axes  :math:`v^+` and :math:`v^-` in :cite:`Seysen22` may be 
obtained as ``MMVector(15, 'I', 2, 3)`` and
``MMVector(15, 'J', 2, 3)``, respectively.

The modulus of a vector can be extended as follows. Assume that 
``v`` is a vector in :math:`\rho_3`, given as an instance of class
|MMVector|. Then ``w = 5 * v`` is a well-defined vector in 
:math:`\rho_{15}`. Vector ``w`` can be constructed as follows:
``V15 = MMV(15); w = V15(5, v)``. Note that the construction
``w = V15(5*v)`` leads to an error, since ``5*v`` is defined
modulo ``3``, but not modulo ``15``. 


Linear order of the basis vectors
..................................

By construction of the representation :math:`\rho_p`, it is most natural 
to use a tuple ``(tag, i0, i1)`` for indexing a basis vector of 
:math:`\rho_p`. There is also a linear order of these entries. Such a 
linear order is required e.g. for expressing a vector in :math:`\rho_p` 
as an array of integers  modulo :math:`p`. Therefore an index given by a 
tuple ``(tag, i0, i1)`` is mapped to linear index ``0 <= i < 196884``. 
Legal standard tuples are listed in the  following table:

.. table:: Conditions for indices ``i0``, ``i1``
          :widths: 25 35 45 

          ================= ==================== ==================== 
          Tag               Condition for ``i0`` Condition for ``i1``
          ================= ==================== ====================
          ``'A'``           ``0 <= i0 < 24``     ``0 <= i1 < 24``
          ``'B', 'C'``      ``0 <= i0 < 24``     ``0 <= i1 < 24``, 
                                                 ``i1 != i0``
          ``'T'``           ``0 <= i0 < 759``    ``0 <= i1 < 64``
          ``'X', 'Y', 'Z'`` ``0 <= i0 < 2048``   ``0 <= i1 < 24``
          ================= ==================== ==================== 

For ``tag = 'A', 'B', 'C'``, we also have 
``(tag, i0, i1) == (tag, i1, i0)``.

The linear order of the tuples ``(tag, i0, i1)`` is as follows:

        offset ``0``:

          | ``(A,0,0), (A,1,1),  ..., (A,23,23)``

        offset ``24``:
 
          | ``(A,1,0),`` 
          | ``(A,2,0),  (A,2,1)``
          | . . .
          | ``(A,23,0),  (A,23,1), ...,  (A,23,22)``

        offset ``300``:
 
          Tuples ``(B,i0,i1)``, same order as 
          tuples ``(A,i0,i1)`` for ``i0 > i1``

        offset ``576``:

          Tuples ``(C,i0,i1)``, same order as  
          tuples ``(A,i0,i1)`` for ``i0 > i1``
           
        offset ``852``:
 
          | ``(T,  0,0),  (T,  0,1), ...,  (T,  0,63)`` 
          | ``(T,  1,0),  (T,  1,1), ...,  (T,  1,63)`` 
          | . . .
          | ``(T,758,0),  (T,758,1), ...,  (T,758,63)`` 

        offset ``49428``:

          | ``(X,   0,0),  (X,   0,1), ...,  (X,   0,23)`` 
          | ``(X,   1,0),  (X,   1,1), ...,  (X,   1,23)`` 
          | . . .
          | ``(X,2047,0),  (X,2047,1), ...,  (X,2047,23)`` 

        offset ``98580``:

          Tuples ``(Z,i0,i1)``, same order as corresponding  
          tuples ``(X,i0,i1)``
       
        offset ``147732``:

          Tuples ``(Y,i0,i1)``, same order as corresponding  
          tuples ``(X,i0,i1)``





Sparse representation of vectors in :math:`\rho_p` 
....................................................

Internally, a vector :math:`\rho_p` is sometimes given in *sparse*
representation. This is useful for sparse vectors containing many
zero entries. In *sparse* representation the vector is given as
an array of unsigned 32-bit integers. Here each entry of the 
array encodes a multiple of the basis vector. Such an entry it 
interpreted as follows:

.. table:: Bit fields in an entry of a sparse representation
          :widths: 28 24 24 24 24 
 
          ========= =========== =========== ========== ========== 
          Component ``tag``     ``i0``      ``i1``     ``factor``
          Bits      ``27 - 25`` ``24 - 14`` ``13 - 8`` ``7 - 0``
          ========= =========== =========== ========== ==========

This corresponds to the tuple ``(factor, tag, i0, i1)``
which denotes a multiple of a basis vector as described in
subsection *Representing a vector as a list of tuples*. 

Tags are mapped to integers as follows:

.. table:: Mapping of integers to tags
          :widths: 14 14 14 14 14 15 15

          ======= ======= ======= ======= ======= ======= ======= 
          ``'A'`` ``'B'`` ``'C'`` ``'T'`` ``'X'`` ``'Z'`` ``'Y'``
           ``1``   ``2``   ``3``   ``4``   ``5``   ``6``   ``7``
          ======= ======= ======= ======= ======= ======= ======= 


Other data types accepted as tags for vectors in :math:`\rho_p` 
................................................................

Some data types are accepted as tags and interprted as described in the
following table. In this case parameters `i0`, `i1` after a tag must not
be set.
    
    .. table:: Legal types for constructing a vector
      :widths: 20 80

      ====================== ============================================= 
      type                   Evaluates to                               
      ====================== ============================================= 
      List of tuples         This evaluates to a vector as described in
                             subsection 
                             *Representing a vector as a list of tuples*.

                             An entry of such a list may also be an 
                             instance of class |MMVector| or a string. 
      ---------------------- --------------------------------------------- 
       class |MMVector|      A deep copy of the given vector is returned.            
      ---------------------- --------------------------------------------- 
       class |XLeech2|       If ``tag`` is of type |XLeech2| then the
                             (possibly negative) basis vector 
                             corresponding to that ``tag`` 
                             in :math:`Q_{x0}` is created.
      ---------------------- --------------------------------------------- 
       ``str``               For an vector ``v`` in ``V`` we have      
                             ``V(str(v)) == v``. 

                             This is helpful for rereading printed 
                             vectors.       
      ====================== ============================================= 



"""
# References in the __docstr__ see file docs/source/references.bib


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import numpy as np
from numbers import Integral
from random import randint
import warnings
import time
from timeit import default_timer
from importlib import import_module
from functools import partial





from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepVector
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepSpace
from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
from mmgroup.structures.mm0_group  import MM0Group, MM0
from mmgroup.structures.abstract_mm_group import is_mmgroup_word
from mmgroup.structures.abstract_mm_rep_space import add_vector

from mmgroup.generators import rand_get_seed
from mmgroup.mm import mm_vector, mm_aux_random_mmv
from mmgroup.mm import mm_aux_zero_mmv, mm_aux_reduce_mmv
from mmgroup.mm import mm_aux_mmv_to_sparse, mm_aux_mmv_add_sparse
from mmgroup.mm import mm_aux_mmv_extract_sparse
from mmgroup.mm import mm_aux_mmv_set_sparse, mm_aux_mmv_add_sparse
from mmgroup.mm import mm_aux_bytes_to_mmv, mm_aux_mmv_to_bytes
from mmgroup.mm import mm_aux_get_mmv, mm_aux_put_mmv
#from mmgroup.mm import mm_rng_make_seed
from mmgroup.mm import INT_BITS, PROTECT_OVERFLOW
from mmgroup.mm import mm_aux_check_mmv
from mmgroup.mm import mm_aux_get_mmv1
from mmgroup.mm import mm_aux_index_extern_to_sparse
from mmgroup.mm import mm_aux_index_sparse_to_extern
from mmgroup.mm import mm_aux_index_sparse_to_leech
from mmgroup.mm import mm_aux_index_sparse_to_leech2

uint_mmv = np.uint32 if INT_BITS == 32 else np.uint64
#standard_seed = mm_rng_make_seed()
standard_mm_group = MM0


TAGS = " ABCTXZY"


######################################################################
# Importing a C wrapper for a specific characteristic 'p'
######################################################################

class MMVectorOps:
    def __init__(self, p):
        self.p = p
        mm = mm_op_modules[p]  # fetch mm_op_modules[p]
        #self.mm = mm_op_modules[p]
                            # if we do this, we cannot pickle vectors
        self.MMV_INTS = mm.MMV_INTS 
        self.op_vector_add = mm.op_vector_add
        self.op_scalar_mul = mm.op_scalar_mul
        self.op_word = mm.op_word
        self.op_compare = mm.op_compare
        self.op_t_A = mm.op_t_A
        self.op_delta = mm.op_delta
        self.op_pi = mm.op_pi
        self.op_t = mm.op_t
        self.op_xi = mm.op_xi
        self.op_xy = mm.op_xy
        self.op_omega = mm.op_omega
        self.op_word_tag_A = mm.op_word_tag_A
        self.mm_vector = partial(mm_vector, p)
        del mm



mm_op_modules = {}
mm_ops = {}
all_characteristics_found = None


def characteristics():
    """Return list of all *characteristics* ``p`` supported"""
    global all_characteristics_found, mm_op_modules, mm_ops
    if all_characteristics_found is None:
        for k in range(2,9):
            p = (1 << k) - 1
            if not p in mm_op_modules:
                try:
                    mm_op_modules[p] = import_module('mmgroup.mm%d' % p)
                except ModuleNotFoundError:
                    pass
            if p in mm_op_modules:
                mm_ops[p] = MMVectorOps(p)
        all_characteristics_found = sorted(mm_op_modules.keys())
    return all_characteristics_found

def get_mm_ops(p):
    try:
        return mm_ops[p]
    except KeyError:
        characteristics()
        if p in mm_ops:
            return mm_ops[p]
        if isinstance(p, Integral):
            err = "Operation mod %s supported in class MMVector"
            raise ValueError(err % p)
        else:
            err = "Modulus for class MMVector must be an integer"
            raise TypeError(err)


######################################################################
# Imports on demand
######################################################################

import_pending = True

def complete_import():
    global mm_op15_eval_A, mm_reduce_2A_axis_type
    global XLeech2, mm_op15_eval_X_count_abs 
    from mmgroup.mm15 import op_eval_A as mm_op15_eval_A
    from mmgroup.mm_reduce import mm_reduce_2A_axis_type 
    from mmgroup.mm15 import op_eval_X_count_abs as mm_op15_eval_X_count_abs
    from mmgroup import XLeech2
    import_pending = False    




#####################################################################
# Modelling a vector of the 196884-dimensional rep of the monster
######################################################################


    

class MMVector(AbstractMmRepVector):
    r"""Models a vector in a representation of the monster group.

    The construction of vectors in the representations :math:`\rho_p`
    of the monster group (using e.g. the constructor of this class)
    is documented at the  beginning of section :ref:`mmrep-label`.
    There a basis vector of that representation is dscribed as a
    tuple ``(tag, i0, i1)``, where ``tag`` is a single capital letter
    and ``i0``, ``i1`` are integers. So we may invoke the constructor
    of this class as follows:

    :param p:   Modulus ``p`` of the representation  :math:`\rho_p`. 

    :param tag: Tag of the unit vector to be constructed. In the
                standard case this is a single capital letter
                describing the type of the basis vector.

    :param i0:  First index in the tuple describing the basis vector.
    
    :param i2:  Second index in the tuple describing the basis vector.
    


    Linear operations can be used for for creating linear combination 
    of unit vectors. This is just the simplest case of the  
    construction of a vector in  :math:`\rho_p`. Section
    :ref:`mmrep-label` contains a large number of further 
    possibilities for creating such vectors.

    Addition and subtraction of vectors in the same space work as 
    usual. This is also the case for scalar multiplication with 
    integers. An element ``g`` of the monster group (given as
    an instance of class |MM|)  operates on a vector ``v`` (given
    as an instance of this class) by right multiplication.

    An entry of a vector ``v`` (given as an instance of this class)
    may be adressed as ``v[tag, i0, i1]``, where ``tag`` is a single 
    letter in the string ``ABCTXYZD`` and ``i0`` and ``i1`` are 
    integers, such that the tuple ``(tag, i0, i1)`` describes a 
    basis vector. Getting and setting entries of a vector is as in
    the ``numpy`` package. Here ``i0`` and ``i1`` may also be 
    slices of integers in the same way as in ``numpy`` arrays. 
    Depending on the ``tag``, the value ``i0`` or ``i1`` may also 
    be a single instance of the appropriate class, as described 
    in the remarks after table :ref:`table-vector-tags` .

    The entries of vector ``v``  also have a linear order as 
    described at the beginnning of section  :ref:`mmrep-label`.
    Here ``v['E', i]`` is the ``i``-th entry in that order.  Index
    ``i`` may also be a slice of integers in the same way as
    in a one-dimensional ``numpy`` array.

    The internal representation of a vector ``v`` in this class
    is not part of the public interface. Use ``v['E']`` to 
    convert ``v`` to an one-dimensional array of ``8``-bit 
    integers of length 196884.
    """
    __slots__ = "p", "data", "ops"
    #group = MM
    def __init__(self, p, tag = 0, i0 = None, i1 = None):
        self.ops = get_mm_ops(p)
        self.p = p
        self.data = mm_vector(p)
        add_vector(self, tag, i0, i1)

    def check(self):
        """Check if the vector is correct

        Raise ValueError if the vector is errorneous.
        """
        self.space.check(self)

       
    def mul_exp(self, g, e = 1, break_g = False):
        """Multiply the vector with ``g ** e`` inplace

        Here ``g`` is an element of the moster group represented
        as an instance of class |MM| and ``e`` is an integer.
        The vector is updated and the updated vector is returned.
        
        Afterwards, the vector contains an attribute ``last_timing``
        containing the run time of this operation in seconds.
        This is useful for benchmarking.

        By default, we try to simplify the expression ``g ** e`` 
        before multiplying it with the vector. If ``break_g`` is 
        set,  we always do ``abs(e)`` muliplications with ``g`` or
        with its inverse.
        """
        return self.space.vector_mul_exp(self, g, e, break_g)
 
    def eval_A(self, v2, e = 0):
        """Internal method, not for public use

        The part of this vector with tag 'A' corresponds to  a
        symmetric 24 times 24 matrix :math:`A`. 

        Let :math:`v_2` be a short vector in the Leech lattice 
        modulo 2 given by  parameter ``v2``, where ``v2`` is an 
        instance of class |XLeech2|. If ``v2`` is an integer then 
        this is converted ``XLeech2(v2)``.

        Then in the Leech lattice shortest preimage :math:`v` of
        :math:`v_2` is determined up to sign and :math:`v A v^\top`
        is determined uniquely.

        In case ``e = 0`` (default) the function returns 
        :math:`v_2 A v_2^\top`. Otherwise the function returns
        :math:`v_2 (A \tau^e) v_2^\top`, where :math:`\tau` is
        the triality element in the monster group.
 
        The current version supports vectors modulo ``p = 15`` only.

        The short Leech lattice vector :math:`v_2` (of norm 4) is 
        scaled to norm 32 as usual, when :math:`v_2` is given in 
        integer coordinates.

        The function raises ValueError if :math:`v_2`  is not
        a short Leech lattice vector or if ``p != 15``.
        """
        if import_pending:
            complete_import()
        if isinstance(v2, XLeech2):
            v2 = v2.value
        v1 = np.zeros(24*4, dtype = np.uint64)
        self.ops.op_t_A(self.data, e % 3, v1)
        if self.p != 15:
            err = "Method eval_A is implemented for vectors mod 15 only"
            raise ValueError(err)
        res = mm_op15_eval_A(v1, v2)
        if res < 0:
            err = "Method eval_A failed on vector in rep of monster"
            raise ValueError(err)
        return res
        

    def count_short(self):
        r"""Count certain entries of the vector

        The function counts the absolute values of all entries 
        of the monomial part of the vector.
   
        Here the monomial part consists of the entries with
        tags 'B', 'C', 'T', 'X'. These entries correspond to 
        the short vectors of the Leech lattice.

        The function returns a tuple of length ``p//2 + 1``, where
        ``p`` is the characteristic of the vector. Entry ``i`` of
        that tuple contains the number of entries of the monomial 
        part of the vector with absolute value ``i``. 

        The returned tuple is invariant under operation of the 
        subgroup :math:`G_{x0}` of the monster group.

        Caution: The function is implemented for characteristic
        ``p = 15`` only!
        """
        if import_pending:
            complete_import()
        if self.p != 15:
            err = "Method supported for characteristic p = 15 only"
            raise ValueError(err)
        a = np.zeros(8, dtype = np.uint32)
        mm_op15_eval_X_count_abs(self.data, a)
        return tuple(a)

    def axis_type(self, e = 0):
        r"""Return axis type if this vector is a 2A axis 

        If this vector is a 2A axis then the function computes the 
        type of the 2A axis. Each 2A axis corresponds uniquely to 
        a 2A involution :math:`x` in the monster. Let :math:`z` be 
        the central involution in the subgroup :math:`G_{x0}` of 
        the monster. Then the type of the 2A axis ``v`` is the class 
        of the product :math:`xz` in the monster. Such a class must 
        be 2A, 2B, 4A, 4B, 4C, 6A, 6C, 6F, 8B, 10A, 10B, or 12C.
   
        In case ``e = 0`` (default) the function returns the type 
        of the axis as a string if this vector is a 2A axis. It 
        may return ``None`` if this vector is not a 2A  axis.

        In case ``e != 0`` the function replaces this vector
        :math:`v` by :math:`v \cdot \tau^e`, where :math:`\tau`
        is the triality element in the monster group.

        The function works for a vector space modulo ``p = 15`` 
        only. It raises ValueError in case ``p != 15`` 

        Caution:

        This is a quick disambiguation of the type of a 2A axis. The 
        function may return any axis type if this is not  a 2A axis.
        """
        if self.p != 15:
            err = "Method supported for characteristic p = 15 only"
            raise ValueError(err)
        e, v = e % 3, self.data
        if e:
            v = np.zeros(24*4, dtype = np.uint64)
            self.ops.op_t_A(self.data, e, v)     
        if import_pending:
            complete_import()
        t = mm_reduce_2A_axis_type(v)
        if t == 0:
            return None
        s = str((t >> 28) & 15) + "?ABCDEFGHIJKLMNO"[(t >> 24) & 15]
        return s
            

######################################################################
# class MMSpace
######################################################################


@singleton
class MMSpace(AbstractMmRepSpace):
    r"""Models a ``196884``-dimensional representation of the monster group 

    This class contains a collection of functions for manipulating
    vectors in the rpresentation :math:`\rho_p` of the monster group.
    Such vectors are instances of class |MMVector|.
    Most of these function are used implicitly in the operators
    applied to these vectors.

    Some of these function may be helpful for the user; so we document 
    them in the *mmgroup API reference*.    
    """
    vector_type = MMVector
    space_name = "MV"

    check_errors = {
        -1: "Bad input value p",
        -2: "A one bit outside a field has been found",
        -3: "A subfield has an illegal nonzero entry at index >= 24",
        -4: "Illegal nonzero diagonal entry", 
        -5: "Symmetric part of vector is not symmetric",
    }



    def __init__(self):
        """Create a 196884-dimensional representation of the monster

        All calculations are done modulo the odd number p
        """
        pass

    @property
    def mm(self, p):
        """Return module object mmgroup.mm<p> for characteristic p"""
        characteristics()
        return mm_op[p]


    #######################################################################
    # Creating vectors 
    #######################################################################

    def zero(self, p):
        """Return the zero vector"""
        return MMVector(p, 0)

    def copy_vector(self, v1):
        assert v1.space == self
        v = MMVector(v1.p, 0)
        np.copyto(v.data, v1.data)
        return v

    def set_rand_uniform(self, v1, seed = None):
        """Return a uniform distributed random vector.

        ``seed`` is a seed for the random generator. The current version 
        supporst the default seed only. Here some random data taken from 
        the operating system and from the clock are entered into the seed.
        """
        seed = rand_get_seed(seed)
        mm_aux_random_mmv(v1.p, v1.data, seed) 
        return v1

    #######################################################################
    # Obtaining and setting components via sparse vectors
    #######################################################################

    def getitems_sparse(self, v1, a_indices):
        """Get items from vector v1

        Here we assert that v1 is a vector of this vector space and
        that 'a_indices' is a one-dimensional numpy array of type
        np.uint32, containing the coordinates to be read from v1.
 
        The function must add the corresponding coordinate to each
        entry of the array 'sparse_items'. All coordinates must be
        nonnegative and < 256. 

        A zero entry in the array 'a_indices' is ignored.
        """
        if len(a_indices):
            mm_aux_mmv_extract_sparse(v1.p, v1.data, a_indices,
                len(a_indices))
        return a_indices 

    def additems_sparse(self, v, a_indices):
        """Add a vector in sparse representation to vector v.

        This method takes a numpy array 'a_indices' of integers of dtype 
        numpy.uint32 containing the description of a vector v2 in sparse 
        representation. It computes 

             v  =  v + v2 .

        Here vector v is a standard vector in this space.
        """
        if len(a_indices):
            mm_aux_mmv_add_sparse(v.p, a_indices, len(a_indices),
                v.data)
        return v

    def setitems_sparse(self, v, a_indices):
        """Set selected components of a vector 

        Arguments 'v' and 'a_indices' are as in method getitems_sparse().
        Here the coordinates of vector 'v' described by 'a_indices' are 
        set to the values given in 'a_indices'. 
        The array 'a_indices' is not changed.
        """
        if len(a_indices):
            mm_aux_mmv_set_sparse(v.p, v.data, a_indices, 
                len(a_indices))
        return v




    #######################################################################
    # Conversion from and to to sparse representation 
    #######################################################################

    def as_sparse(self, v1):
        """Yet to be documented!!

        """
        sp = np.zeros(196884, dtype = np.uint32)
        length = mm_aux_mmv_to_sparse(v1.p, v1.data, sp)
        return sp[:length]


    #######################################################################
    # Vector operations 
    #######################################################################


    def iadd(self, v1, v2):
        if v1.p == v2.p:
            v1.ops.op_vector_add(v1.data, v2.data)
            return v1
        else:
            err = "Cannot add vectors modulo differnt numbers"
            raise ValueError(err)
 
    def imul_scalar(self, v1, a):
        v1.ops.op_scalar_mul(a % v1.p, v1.data)
        return v1
           
    #######################################################################
    # Group operation 
    #######################################################################

    def imul_group_word(self, v1, g):
        """Return product v1 * g of vector v1 and group word g.

        v1 may be destroyed.

        This method is called for elements v1 of the space
        'self' and for elements g of the group 'self.group' only.
        """
        work =  mm_vector(v1.p)
        if isinstance(g, AbstractMMGroupWord):
            a = g.mmdata
            v1.ops.op_word(v1.data, a, len(a), 1, work)
            return v1
        err = "Multiplicator for MM vector must be int or in MM group"   
        raise TypeError(err) 
 

    def vector_mul_exp(self, v1, g, e, break_g = False):
        """Compute product v1 * g**e of vector v1 and group word g.

        Here v1 is a vector in this space, e is an integer, g is a 
        group element, and  v1 is replaced by v1 * g**e.

        This method should be  called for elements v1 of the space
        'self' and for elements g of the group 'self.group' only.

        If break_g is True, each factor g is multiplied with v1
        separately. Otherwise, the expression  g**e  may be
        optimized. This option is mainly for benchmarking.

        After applying this function to vecter v1, the vector
        v1 has an attribute v1.last_timing containing the run
        time of the C part of this operation in seconds.
        """
        work = mm_vector(v1.p)
        assert v1.space == self
        assert -1 << 31 < e < 1 << 31
        assert isinstance(g, MM0) 
        length = g.length
        if break_g:
            g._extend(length + 1)
            g._data[length] = 0x70000000
            length += 1
        t_start = time.perf_counter()
        #t_start = default_timer()
        v1.ops.op_word(v1.data, g._data, length, e, work)
        v1.last_timing = time.perf_counter() - t_start
        #v1.last_timing = default_timer() - t_start
        return v1  


    #######################################################################
    # Checking equality
    #######################################################################

    def equal_vectors(self, v1, v2):
        """Return True iff vectors v1 and v2 are equal 

        This method is called for elements v1 and v2 of the space
        'self' only.
        """
        if v1.p == v2.p:
            return not v1.ops.op_compare(v1.data, v2.data) 
        return False

    #######################################################################
    # Conversion from and to byte format
    #######################################################################

    def as_bytes(self, v1):
        """Return vector 'self' as a byte array

        The result is a numpy array with dtype = uint8 and
        shape = (196884,).
        """
        b = np.zeros(196884, dtype = np.uint8)
        mm_aux_mmv_to_bytes(v1.p, v1.data, b)
        return b

    def from_bytes(self, p, b):
        """Construct a vector from a byte array

        Here ``b`` is an array-like object representing a 
        one-dimensional array of ``196884`` integers as in the
        ``numpy`` package. These integers are taken modulo the
        characteristic ``p`` of the space. 

        Array ``b`` represents a vector in this space in linear order
        as described method ``tuple_to_index``. 

        The function returns the vector given by the array ``b`` in
        this vector space as an instance of class |MMVector|.
        """
        b = np.array(b, dtype = np.int32)
        if len(b.shape) != 1:
            raise TypeError("Bad shape of byte data vector")
        if len(b) != 196884:
            raise TypeError("Bad length of byte data vector")
        b = np.array(b % p, dtype = np.uint8)
        v =self.zero(p)
        mm_aux_bytes_to_mmv(p, b, v.data)
        return v

 
        
    #######################################################################
    #  Checking and reducing a vector
    #######################################################################

    def check(self, v1):
        """Check the vector 'self'.

        Raise ValueError if an error is found in vector 'self'.
        """
        if len(v1.data) != v1.ops.MMV_INTS + 1:
            err = "MM vector has wrong length"
            raise MemoryError(err)   
        if v1.data[-1] != PROTECT_OVERFLOW:
            err = "Buffer overflow in MM vector detected"
            raise MemoryError(err)
        result = mm_aux_check_mmv(v1.p, v1.data)
        #print("check result is", result)
        if not result:
            return
        try:
            err = self.check_errors[result]
        except:
            err = "Unknown error %d in MM vector" % result
        print("\n%s!\n" % err)
        raise ValueError("Error in MM vector")
 
    def reduce(self, v1):
        """Convert vector v1 to a unique reduced form"""
        mm_aux_reduce_mmv(v1.p, v1.data)
        return v1

    #######################################################################
    # Conversion between tags and indices
    #######################################################################

    @classmethod
    def tuple_to_index(cls, tag, i0 = -1, i1 = -1):
        r"""Convert tuple ``(tag, i0, i1)`` to a linear index


        Remarks:

        The tuple ``('D', i0)`` is accepted as a shorthand for 
        ``('A', i0,i0)``.

        A tuple ``('E', i0)`` means a linear index ``i0``,  i.e.
        this method returns ``i0`` on input ``('E', i0)``, for
        ``0 <= i0 < 196884``.

        If ``tag`` is an instance of class |MMVector|, which is
        a nonzero multiple of a basis vector, then the linear index 
        corresponding to that basis vector is returned.        
        """
        i = 0
        if isinstance(tag, str) and len(tag) == 1:
            t = TAGS.find(tag)
            if  t >= 1 and 0 <= i0 < 2048 and 0 <= i1 < 64:
                i = (t << 25) + (i0 << 14) + (i1 << 8) 
            elif  tag == "E" and 0 <= i0 < 196884:
                return i0
            elif tag == "D" and 0 <= i0 < 24:
                i = (1 << 25) + (i0 << 14) + (i0 << 8) 
        elif isinstance(tag, MMVector):
            sp = tag.as_sparse()
            if len(sp) == 1:
                i = sp[0] 
            else:
                err = "MM vector is not multiple of basis vector"
                raise ValueError(err)
        else:
            raise TypeError("Cannot convert object to MM vector index")
        i_ext = mm_aux_index_sparse_to_extern(i)
        if 0 <= i_ext < 196884:
            return i_ext
        err = "Could not convert tuple with tag %s to MM vector index"
        raise ValueError(err % tag)

 
    @classmethod
    def index_to_tuple(cls, index):
        """Convert linear index to tuple ``(tag, i0, i1)``

        This method reverses the effect of method ``tuple_to_index``.
        Given a linear index ``0 <= i < 196884`` for a basis
        vector, the function returns that index as a tuple
        ``(tag, i0, i1)`` with ``tags`` one letter of the
        string ``"ABCTXYZ"`` and integers ``i0``, ``i1``.

        See method ``tuple_to_index`` for details.

        If ``index`` is an instance of class |MMVector|, which is
        a nonzero multiple of a basis vector, then the index of that 
        basis vector is taken.        
        """
        if isinstance(index, Integral): 
            if  0 <= i < 196884:
                i = mm_aux_index_extern_to_sparse(index)
                return TAGS[i >> 25], (i >> 14) & 0x7ff, (i >> 8) & 0x3f
            else:
                raise ValueError("MM vector index out of range")
        elif isinstance(index, MMVector):
            sp = index.as_sparse()
            if len(sp) == 1:
                i = sp[0]
                return TAGS[i >> 25], (i >> 14) & 0x7ff, (i >> 8) & 0x3f
            else:
                raise ValueError("MM vector is not multiple of basis vector")
        else:    
            raise TypeError("Cannot convert object to MM index tuple")


    #######################################################################
    # Conversion to short Leech lattice vector
    #######################################################################


    @staticmethod
    def index_to_sparse(tag, i0 = -1, i1 = -1):
        r"""Auxiliary method for index_to_short

        Convert a tagged tuple ``(tag, i0, i1)`` to a sparse 
        index. That tuple must refer to an index describing a
        short Leech lattice vector.
        See method ``index_to_short`` for details. 
        """
        i = 0x0
        if isinstance(tag, Integral) and 300 <= tag < 98580:
            i = mm_aux_index_extern_to_sparse(tag)
        elif isinstance(tag, str) and len(tag) == 1:
            t = TAGS.find(tag)
            if  t >= 1 and 0 <= i0 < 2048 and 0 <= i1 < 64:
                i = (t << 25) + (i0 << 14) + (i1 << 8) 
            elif  tag == "E" and 300 <= i0 < 98580:
                i = mm_aux_index_extern_to_sparse(i0)
            elif tag == "D" and 0 <= i0 < 24:
                i = (1 << 25) + (i0 << 14) + (i0 << 8) 
        elif isinstance(tag, MMVector):
            sp = tag.as_sparse()
            if len(sp) == 1:
                i = sp[0]
        else:    
            err = "Cannot convert object to short Leech lattice vector"
            raise TypeError(err)
        return i

    @staticmethod
    def index_to_short(tag, i0 = -1, i1 = -1):
        r"""Convert index to a short Leech lattice vector

        If ``tag`` is an integer, this is interpreted a linear index 
        for a basis vector in the representation :math:`\rho_p` as in 
        method ``tuple_to_index``. Otherwise the tuple ``(tag, i0, i1)`` 
        is interpreted a standard index for such a basis vector.
        If ``tag`` is an instance of class |MMVector|, which is
        a nonzero multiple of a basis vector, then that basis vector 
        is taken.
 
        Some but not all of these basis vectors correspond to short
        vector in the Leech lattice up to sign. If this is the
        case then the function returns a short vector of the Leech
        lattice as a numpy array of signed ``32``-bit integers of 
        length ``24``. The norm of the returned vector is ``32``.

        The function raises ValueError if the basis vector in 
        :math:`\rho_p` does not correspond to a short vector.

        The sign of the short vector is undefined, but two calls
        referring to the same basis vector return the same short
        vector.   
        """
        i = MMSpace.index_to_sparse(tag, i0, i1)
        v = np.zeros(24, dtype = np.int32)
        if mm_aux_index_sparse_to_leech(i, v) == 0:
            return v          
        err = "Vector does not map to short Leech lattice vector"
        raise ValueError(err)

    @staticmethod
    def index_to_short_mod2(tag, i0 = -1, i1 = -1):
        r"""Convert index to a short Leech lattice vector modulo 2

        The tuple ``(tag, i0, i1)`` is interpreted as a basis vector
        as in method ``index_to_short``.
 
        Some but not all of these basis vectors correspond to short
        vector in the Leech lattice up to sign. If this is the
        case then the function returns a short vector of the Leech
        lattice modulo 2 as an integer. 

        The function raises ValueError if the basis vector in 
        :math:`\rho_p` does not correspond to a short vector.
        """
        i = MMSpace.index_to_sparse(tag, i0, i1)
        i2 = mm_aux_index_sparse_to_leech2(i)
        if i2 != 0:
            return i2          
        err = "Vector does not map to short Leech lattice vector"
        raise ValueError(err)


StdMMSpace = MMSpace()
MMVector.space = StdMMSpace



def MMV(p):
   r"""Return an object correponding to the space :math:`\rho_p`

   Here characteristic ``p`` is as in the constructor of
   class |MM|. Thus the sequence

   .. code-block:: python

     >>> V = MMV(p)  
     >>> v = V(tag, i0, i1)

   returns the same vector in :math:`\rho_p` as the 
   statement ``MMVector(p, tag, i0, i1)``. 

   Function ``characteristics`` in module ``mmgroup`` returns
   the list of legal values for the characteristic ``p``.

   Technically, ``MMV(p)`` is the partial application of
   the constructor of class |MMVector| to the number ``p``. 
   """
   return partial(MMVector, p)



def order_vector():
   r"""Return the precomputed order vector

   The function returns the precomputed order vector :math:`v_1`
   (which is an element of the representation  :math:`\rho_{15}`
   of the monster) as an instance of class |MMVector|.

   The properties required for such a vector :math:`v_1`
   are defined in :cite:`Seysen22`.
   """
   from mmgroup.mm_reduce import mm_order_load_vector
   v = MMVector(15)
   mm_order_load_vector(v.data)
   return v






r"""Generation of  C code dealing with the Mathieu group Mat24

Generating the ``mmgroup.mat24`` extension
..........................................

Function ``mat24_make_c_code()`` generates C code for basic computations
in the Golay code, its cocode, and the Mathieu group Mat24. It also
generates code for computations in the Parker loop and in its 
automorphism group. 

The generated C modules are used in the python extension 
``mmgroup.mat24``. The functions used by that extension are contained 
in a shared library with name ``mmgroup_mat24.dll``. The reason for
creating such a shared library is that that these functions are also
called by C functions written for other python extensions.

We use the C code generation mechanism in class 
``generate_c.TableGenerator``. Here a .c file  and a .h file is
crreated from file ``mat24_functions.ske`` in subdirectory
``src/mmgroup/dev/mat24``. The .ske file is like a .c file, but 
augmented with some code generation statements for entering tables 
and automatically generated code into the .c file to be generated. 
This .ske file may also have statements for automatically 
generating a .h file declaring the exported functions.  

We create an instance ``tg`` of class ``TableGenerator`` for 
generating the .c files. The table generator ``tg`` takes two
dictionaries ``tables`` and ``directives`` as arguments. These
dictionaries provide user-defined tables and directives for the
code generator. Class  ``Mat24`` in module
``mmgroup.dev.mat24.mat24_ref`` has methods ``tables()`` 
and ``directives()`` creating the required tables and directives.


Generating the ``mmgroup.generators`` extension
.............................................

Function ``generators_make_c_code`` generates C code for computing 
the monomial part of the operation of the elements :math`\xi` and
:math`\xi^2`of the monster group. These C functions are used for
computing (rather large) tables required for the implmentation of
the functions that compute the operation :math`\xi` and :math`\xi^2` 
on a representation of the monster.

The generation of the ``mmgroup.generators`` extension is similar
to the generation of the ``mmgroup.mat24`` extension. Here the
list of .ske file is given in the list GENERATORS_C_FILES.
For each file in that list a C file is created. 

A common header with name given by H_GENERATORS_NAME is created 
from all these .ske files, prependend by the header files in
the list GENERATORS_H_FILES. A .pxd file with name
PXD_GENERATORS_NAME is created from that header file. That .pxd
file will also contain the declarations in the string
PXD_DECLARATIONS.

All input files are read fom the directory SKE_DIR.


 
Location of the output files
............................

The location of the generated output files is controlled by certain
variables in module config.py. Each of these variables specifies the
name of a directory.

Files with extension .c, .h go to the directory ``C_DIR``. Files with
extension .pxd, .pxi, .pyx go to the directory ``PXD_DIR``. 
"""



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import types
import sys
import re
import os
import subprocess
import shutil



from config import SRC_DIR, DEV_DIR,  C_DIR, PXD_DIR
from config import REAL_SRC_DIR 
sys.path.append(REAL_SRC_DIR)

from mmgroup.dev.mat24.mat24_ref import Mat24
from mmgroup.dev.generators.gen_xi_ref import GenXi
from mmgroup.dev.generators.gen_leech_reduce_n import GenLeechReduceY
from mmgroup.dev.generators.gen_cocode_short import ShortCocodeTables
from mmgroup.generate_c import TableGenerator, make_doc
from mmgroup.generate_c import pxd_to_pyx


########################################################################
# Generate mat24_functions.c
########################################################################

pxd_declarations = """
from libc.stdint cimport uint32_t, uint16_t, uint8_t

"""



def mat24_make_c_code():
    """Create .c and .h file with the functionality of class Mat24

    The input of this function is the file MAT24_C_FILE.ske that 
    contains a (much faster) C version of the functions in class
    mmgroup.dev.mat24.mat24_ref.Mat24.

    The functions in the .ske file make use of the tables that have 
    been generated for this module and also of some functions for
    generating C code automatically. An example where generating
    code automatically makes sense is the matrix multiplication with
    a constant bit matrix.
     
    The code generating process is described in class TableGenerator
    in module make_c_tables.  
    """ 
    print("Creating C source from file mat24_functions.ske\n")
    MAT24_C_FILE = "mat24_functions"
    SKE_DIR = os.path.join(DEV_DIR, "mat24")
    # The follwing two tables can't easily be computed earlier
    Mat24.tables["Mat24_doc_basis"] = Mat24.str_basis()
    generator = TableGenerator(Mat24.tables, Mat24.directives)
    f = os.path.join(SKE_DIR, MAT24_C_FILE)
    path_ = os.path.join(C_DIR, MAT24_C_FILE)  
    #print("pwd", os.getcwd())
    #print(os.path.realpath(path_ + ".c"))
    generator.generate(f + ".ske", path_ + ".c", path_ + ".h")
    ## generator.export_tables(file_name = "mat24_export.py")

    generator.generate_pxd(
        os.path.join(PXD_DIR, MAT24_C_FILE + ".pxd"), 
        MAT24_C_FILE + ".h",  
        pxd_declarations   
    )
    print("C files for extension mat24 have been created" )
    


########################################################################
# Generate c files for module 'generators'
########################################################################

SKE_DIR =  os.path.join(DEV_DIR, "generators")

GENERATORS_C_FILES = [
    "gen_xi_functions",
    "mm_group_n",
    "gen_leech",
    "gen_leech_type",
    "gen_leech3",
    "gen_leech_reduce",
    "gen_leech_reduce_n",
    "gen_random",
]


GENERATORS_H_START = """
// %%GEN h
#ifndef MMGROUP_GENERATORS_H
#define MMGROUP_GENERATORS_H
// %%GEN c

"""

GENERATORS_H_END = """
// %%GEN h
#endif // ifndef MMGROUP_GENERATORS_H
// %%GEN c
"""

GENERATORS_H_FILES = [
    GENERATORS_H_START,
    "mmgroup_generators.h",
]

GENERATORS_TABLE_CLASSES = [
     GenXi, GenLeechReduceY, ShortCocodeTables
]

H_GENERATORS_NAME = "mmgroup_generators.h"
PXD_GENERATORS_NAME = "generators.pxd"
PXI_GENERATORS_NAME = "generators.pxi"

PXD_DECLARATIONS = """

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libc.stdint cimport int64_t, int32_t

"""


def generators_make_c_code():
    """Create .c and .h file with the functionality of class Mat24Xi

    """
    print("Creating C sources for the 'generators' extension\n")

    # Setp table and directives for code generation
    GenXi.tables["GenXi_doc"] = GenXi # can't do this earlier
    tables = {}
    directives = {}
    for table_class in GENERATORS_TABLE_CLASSES:
        table_instance = table_class()
        tables.update(table_instance.tables)
        directives.update(table_instance.directives)
    print(tables.keys())
    tg = TableGenerator(tables, directives)

    # Generate c files
    all_ske_files = [os.path.join(SKE_DIR, name) 
        for name in GENERATORS_H_FILES]
    for name in GENERATORS_C_FILES:
        ske_file = name + ".ske"
        ske_path = os.path.join(SKE_DIR, ske_file)
        c_file = name + ".c"
        c_path = os.path.join(C_DIR, c_file)
        print("Creating %s from %s" % (c_file, ske_file))
        tg.generate(ske_path, c_path)
        all_ske_files.append(ske_path)

    # generate .h file
    all_ske_files.append(GENERATORS_H_END)
    h_file =  H_GENERATORS_NAME
    h_path =  os.path.join(C_DIR, h_file)
    pxd_file =  PXD_GENERATORS_NAME
    print("Creating %s from previous .ske files" % h_file)
    tg.generate(all_ske_files, None, h_path)

    # generate .pxd file
    tg.generate_pxd(
        os.path.join(PXD_DIR, PXD_GENERATORS_NAME), 
        h_file, 
        PXD_DECLARATIONS
    )
    print("C files for extension 'generators' have been created" )
        
    # generate .pxi file
    def pxi_comment(text, f):
        print("\n" + "#"*70 + "\n### %s\n" % text + "#"*70 + "\n\n",
        file=f
    )
    f_pxi = open(os.path.join(PXD_DIR, PXI_GENERATORS_NAME), "wt")
    pxi_comment(
        "Wrappers for C functions from file %s" % PXD_GENERATORS_NAME, 
        f_pxi
    )
    print(PXD_DECLARATIONS, file = f_pxi)
    pxi_content = pxd_to_pyx(
        os.path.join(PXD_DIR, PXD_GENERATORS_NAME),
        os.path.split(PXD_GENERATORS_NAME)[0],
        select = True
    )
    print(pxi_content, file = f_pxi)
    f_pxi.close()

 
########################################################################
# Main program
########################################################################

if __name__ == "__main__":
    mat24_make_c_code()    
    generators_make_c_code()    
    


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
import math

import pytest

from mmgroup.bitfunctions import bit_mat_random, bit_mat_inverse
from mmgroup.bitfunctions import bit_mat_mul, bit_mat_rank
from mmgroup.bitfunctions import lrange, lmap, bitlen, v2
from mmgroup.bitfunctions import bit_mat_orthogonal_complement
from mmgroup.bitfunctions import bit_mat_transpose
from mmgroup.bitfunctions import bit_mat_iszero


 
def inverse_testcases():
    testcases = [  [3,4,6,12], [2,3],
    ]
    NTESTS = 100
    for a in testcases:
         yield a
    for j in range(NTESTS):
        for n in lrange(1,13) + [24]: 
            a = bit_mat_random(n, n)
            yield a
    for j in range(200): yield  bit_mat_random(10, 10)


@pytest.mark.bitfunc
def test_inverse():
    print( "Testing matrix inversion ..." )
    MAXDIM_INVERSE_TEST = 20
    NSIGMA = 4.0
    elementary_testcases =  [  ([3,2], [3,2]) , ([2,4,1], [4,1,2]) ]
    for a, b in elementary_testcases:
        assert  bit_mat_inverse(a) == b
        if a != b: assert  bit_mat_inverse(b) == a


    n_cases = [0] *(MAXDIM_INVERSE_TEST+1) 
    n_ok = [0.0] *(MAXDIM_INVERSE_TEST+1) 
    for a in inverse_testcases():
        ok, n = False, len(a)
        if n > MAXDIM_INVERSE_TEST:
            continue
        n_cases[n] += 1
        try:
            inv = bit_mat_inverse(a)
            ok = True
            n_ok[n] += 1
        except:
            pass
        if ok:
            unit = [1 << i for i in range(n)]
            prod = bit_mat_mul(a, inv) 
            assert prod == unit, (lmap(hex,a), lmap(hex,inv), lmap(hex,prod))
    cases_displayed = [10]
    for n in range(1,MAXDIM_INVERSE_TEST):
        if n_cases[n] > 10:
            import math
            N, N_ok = n_cases[n], n_ok[n]
            p = 1.0
            for i in range(1,n+1): p *= 1.0 - 0.5**i
            mu =  N * p
            sigma = math.sqrt(N * p * (1.0 - p))
            if n in cases_displayed:
                print("Dim. %d, %d cases, ratio of invertibles: %.3f, expected: %.3f+-%.3f"  %(
                     n, N, N_ok/N, p, sigma/N))
            assert   abs(mu - N_ok) < NSIGMA *sigma, (n,p,mu,sigma,N,N_ok)
    print("Matrix inversion test passed")



def orthogonal_complement_testcases():
    testcases = [  ([6],5), ([3,4,6,12],None), ([2,3],None)
    ]
    for a, r in testcases:
         yield a, r
     
    for m in range(1,20):
        for n in range(1,20):
            a = bit_mat_random(m, n)
            yield a, n
   

@pytest.mark.bitfunc
def test_orthogonal_complement():
    for a, ncols in orthogonal_complement_testcases():
        ra = bit_mat_rank(a)
        ao = bit_mat_orthogonal_complement(a, ncols)
        prod = bit_mat_mul(a, bit_mat_transpose(ao))
        assert bit_mat_iszero(prod) , (a, ncols, ao)
        rao = bit_mat_rank(ao) 
        if ncols:
            assert ra + rao == ncols, (ra, rao, ncols, a, ao)
    print( "Test of othogonal complement computation passed" )


def v2_testcases():
    for i in lrange(32) + lrange(32,70,3):
         yield (1 << i)
         if i: yield (1 << i) - 1


def v2_old(x):
    v = 0
    assert x
    while x & 1 == 0:
        x, v = x >> 1, v + 1
    return v

@pytest.mark.bitfunc
def test_v2():
    for i in lrange(300)+[100000]:
        assert bitlen(1<<i) == i+1           
        assert bitlen((1<<i)-1) == i   
    for i in  v2_testcases():
         assert v2(i) == v2_old(i), (hex(i), v2(i), v2_old(i))
         pass


        



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
import collections
import re
import warnings
from numbers import Integral
import numpy as np
from random import choice

import pytest

from mmgroup.tests.groups.mgroup_n import MGroupN
from mmgroup.tests.groups.mgroup_n import n_rules, n_inv


###########################################################################
# Test functions
###########################################################################

MIN_VERBOSE = 2




RAND_TAGS = "dpxyt" * 10
N_TESTS = 200

TEST_PAR = [
   (N_TESTS, RAND_TAGS)
]

def g_sample(g, tags):
    atoms = [(choice(tags), 'r') for i in range(7)]
    return g(atoms)


@pytest.mark.auto_group
@pytest.mark.parametrize("n_tests, rand_tags", TEST_PAR )
def test_mgroupn(n_tests, rand_tags, verbose = 0):
    global n_rules, n_inv
    n_rules =  n_inv = 0
    g = MGroupN()
    x1 = g('x', 0x1234)
    x2 = g('x', 0x1fe7)
    y1 = g('y', 0x1fe1)
    y2 = g('y', 0x1e31)
    p = g([('d', 0x3b5), ('p', 123423458)])
    print("x1*x2",  x1, x2,  (x1 * x2)**(-3) )
    print( g_sample(g, rand_tags).reduce() )
    for i in range(n_tests):
        #w1, w2, w3 = g(rand_tags), g(rand_tags), g(rand_tags)
        w1, w2 = g_sample(g, rand_tags), g_sample(g, rand_tags)
        w3 = g_sample(g, rand_tags) 
        assert g(str(w1)) == w1, (g(str(w1)), w1)
        assert g(w1.as_tuples()) == w1
        l_assoc = ((w1 * w2) * w3).reduce()
        r_assoc = w1 * (w2 * w3)
        ok = l_assoc == r_assoc
        if verbose or not ok or i < MIN_VERBOSE:
            print("\nTest associative law")
            print("w1 =", w1)			
            print("w2 =", w2)			
            print("w3 =", w3)			
            print("(w1*w2)*w3 =", l_assoc )
            if not ok:			
                print("w1*(w2*w3= =", w1 * r_assoc )			
                raise ValueError("Test of accociative law Failed")
        a = (w1 * w2 * w3)**(-1) 
        b =  w3**(-1) * w2**(-1) * w1**(-1)
        ok = a == b
        if verbose or not ok or i < MIN_VERBOSE:
            print("\nTest Inversion of w1 * w2 * w3")
            print("w1 =", w1)			
            print("w2 =", w2)
            print("w3 =", w3)
            print("a = (w1 * w2 * w3)**(-1)")			
            print("a =", a)			
            print("b = w3**(-1) * w2**(-1) * w1**(-1)")			
            print("b =", b)
            if not ok:			
                print("a * b**(-1)) =", a * b**(-1)) 
                print("lengths of a and b",  len(a), len(b))
                print("atom comparison", [ x == y for x, y in zip(a, b)] )
                raise ValueError("Inversion Failed")
    #print("%d rules, %d inversions" % (n_rules, n_inv))





#if __name__ == "__main__":
#    test_mgroupn()





"""Test auxiliary C functions in generated file  mat24_functions.c

These functions are wrapped in class mat24fast.Mat24 and used for the
representation of the monster group.
"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys 
from random import shuffle, randint
import pytest

from mmgroup import mat24
Mat24 = mat24

def op_all_autpl_testdata():
    def to_autpl(c, perm_num):
        p = Mat24.m24num_to_perm(perm_num)        
        m = Mat24.perm_to_autpl(c, p) 
        assert (m[11] >> 12) & 1 == (c >> 11) & 1
        return m
    testdata = [ 
        (1, 0, [1 << i for i in range(11)]),
        (3, 17, [1 << i for i in range(11)]),
    ]
    for d, n,  vlist in  testdata:
        yield to_autpl(d, n),  vlist
    for i in range(500):
        m = to_autpl(randint(0, Mat24.MAT24_ORDER - 1), randint(0,0xfff))
        yield (m ,  [randint(0, 0x7ff) for i in range(20)])





@pytest.mark.mat24
def test_op_all_autpl():
    print("Testing function op_all_autpl()...", end="")
    for m, vlist in op_all_autpl_testdata():
        v_all = Mat24.op_all_autpl(m)
        odd = (m[11] & 0x1000) >> 12
        for v in vlist:
            w0 = v_all[v]
            signs = [(w0 >> i) & 1 for i in (14,13,12)]
            w = (w0 & 0xfff)
            im_v = Mat24.op_ploop_autpl(v, m) 
            assert (im_v & 0xfff) ==  w, (v, im_v, hex(w0))
            assert im_v >> 12 == signs[1], (v, im_v, signs, hex(w0))
            assert ((im_v >> 12) ^ (im_v >> 11)) & 1 == signs[0], (
                  v, im_v, signs, hex(w0))
            pwr = signs[1] ^ signs[2]
            weight = Mat24.gcode_weight(im_v)  & 1
            assert pwr == weight & odd, (v, im_v, signs, weight, hex(w0))
    print(" passed")
  

@pytest.mark.mat24
def test_autpl_cocode():
    print("Testing function op_all_cocode()...", end="")
    for i in range(100):
        c = randint(0,0xfff)
        m = Mat24.perm_to_autpl(c, list(range(24))) 
        v_all = Mat24.op_all_autpl(m)
        v_c = Mat24.op_all_cocode(c)
        v_all = [(x >> 12) & 7 for x in  v_all]
        v_c =  [x & 7 for x in  v_c]
        if v_all != v_c:
            for j in range(len(v_all)):
                if v_all[j] != v_c[j]:
                    print("cocode = %s, v_all[%s] = %s, v_c[%s] = %s" % 
                        tuple(map(hex, (c, j, v_all[j], j, v_c[j]))))
                    print(v_all[:32])
                    print(v_c[:32])
                    err = "Testing of function op_all_cocode() failed"
                    raise ValueError(err) 
                    break




r"""We deal with a certain group of automorphisms of the Parker loop.

A *standard automorphism* is an automorphism of the Parker loop 
:math:`\mathcal{P}`  which maps to an automorphism of the Golay 
code :math:`\mathcal{C}` when  factoring out the elements 
:math:`\{1,-1\}` of :math:`\mathcal{P}`. Let
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` be the 
group of standard automorphisms of  :math:`\mathcal{P}`. We 
embed :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`
into :math:`{\mathbb{M}}` as in :cite:`Con85`. 
In ATLAS notation, see :cite:`Atlas`, that group has structure

.. math::
      {{\rm Aut}_{{\rm St}} \mathcal{P}} =
         2^{12} . M_{24}\,  .
         

This means that :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` has an
elementary Abelian normal subgroup of order  :math:`2^{12}` with
factor group  :math:`M_{24}`. That group extension does not
split. Here the group :math:`2^{12}` is isomorphic
to the Golay cocode :math:`\mathcal{C}^*`  and  :math:`M_{24}`
is the Mathieu group acting on :math:`\mathbb{F}_2^{24}`
by permuting the basis vectors. :math:`M_{24}` is the
automorphism group of the Golay code  :math:`\mathcal{C}` .

The automorphisms in the subgroup :math:`\mathcal{C}^*` of
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` are called 
*diagonal* automorphisms. A diagonal automorphism  
:math:`d \in \mathcal{C}^*` maps an element :math:`a` of  
:math:`\mathcal{P}` to :math:`(-1)^s \cdot a` with 
:math:`s = |a \, \& \, d|`. 
Here :math:`\&` means the bitwise ``and`` operation of bit 
vectors and :math:`|.|` means the bit weight of a bit vector.


Since the extension :math:`2^{12} . M_{24}` does not split,
there is no canonical embedding of :math:`M_{24}` into 
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. For practical 
purposes we need such an embedding. We choose a basis
:math:`b_1,\ldots,b_{11}` of the Golay code :math:`\mathcal{C}`,
see section :ref:`basis-golay-label`. Let :math:`a_i` be
the positive element of the Parker loop :math:`\mathcal{P}`
corresponding  to  :math:`b_i`. For any 
:math:`\tilde{g} \in M_{24}` we let :math:`g` be the unique 
element of :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` with
:math:`g \mathcal{C}^* = \tilde{g}` 
that maps all elements :math:`a_1,\ldots,a_{11}` of
:math:`\mathcal{P}` to positive elements of :math:`\mathcal{P}`.
We call :math:`g` the *standard representative* of 
:math:`\tilde{g}` in :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. 


Elements of :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` are 
modelled as instances of class |AutPl|. If ``d`` is an instance
of class |Cocode| representing an element of  :math:`\mathcal{C}^*`
then ``AutPL(d)`` is the corresponding diagonal automorphism.

A permutation :math:`p \in M_{24}` can be represented as a
list ``l_p = [p(0),...,p(23)]``. If ``l_p`` is such a list
then ``AutPL(l_p)`` is the standard representative of ``p``
in  :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. Here an error
occurs if the mapping from ``i`` to ``p(i)`` is not in
:math:`M_{24}`. A permutation in :math:`M_{24}` can also be 
given as a mapping from a subset of :math:`\{0,\ldots,23\}`
to  :math:`\{0,\ldots,23\}`, provided that this mapping
extends to a unique element of :math:`M_{24}`. For details,
see class  |AutPl|. 


The group :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` operates
naturally on :math:`\mathcal{P}` by right multiplication.
It also operates on  :math:`\mathbb{F}_2^{24}`, 
:math:`\mathcal{C}`,  and :math:`\mathcal{C}^*` by right 
multiplication. Here the basis vectors of 
:math:`\mathbb{F}_2^{24}` are permuted by :math:`M_{24}`;
so the kernel of that operation is the subgroup
:math:`\mathcal{C}^*` of 
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`.

We order the elements  :math:`p` of :math:`M_{24}` in lexicographic
order based on the lists ``l_p = [p(0),...,p(23)]``. We number the
``244823040`` elements of :math:`M_{24}` in that order from ``0`` 
to ``244823039``.  Thus the number ``0`` is assigned to the neutral
element of  :math:`M_{24}`, and the number ``244823039`` is 
assigned to the permutation ``p`` with ``l_p`` =


    :math:`\small [ 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12, 
    11, 10, 9, 8, 0, 1, 2, 3, 4, 5, 6, 7 ]`

These numbers are used for addressing the corresponding standard 
representatives in the subgroup 
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` of the
monster :math:`\mathbb{M}`. 

The easiest way to create an element of :math:`M_{24}` is to
map an *umbral heptad* to another umbral heptad. Here an
umbral heptad is a set of ``7`` elements of 
:math:`\{0,\ldots,23\}` such that precisely ``6`` of these
elements are contained in an octad, see :cite:`CS99`, 
Chapter 10. Then the ``6`` elements in one octad must map 
to the ``6`` elements in the other octad.

The set ``[0,1,2,3,4,5,8]`` is an umbral heptad with ``8``
not contained in the octad. We may create the standard 
representative (in :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`)
of a random element of :math:`M_{24}` as follows:


.. code-block:: python

  from random import sample
  from mmgroup import Octad, AutPL
  # Create a random octad
  o = Octad("r")
  # Select 6 random elements from that octad
  hexad = sample(o.bit_list, 6)
  # Select one random element not in that octad
  extra = sample((~o).bit_list, 1)
  # Create umbral heptad from these random elements
  heptad = hexad + extra
  # Create a mapping from [0,1,2,3,4,5,8] to that heptad 
  hmap = zip([0,1,2,3,4,5,8], heptad)
  # Create a permutation in the Mathieu group from that mapping
  aut = AutPL(0, hmap) 
  # 'aut' is the standard representative of the permutation
  # Show 'aut' as a permutation in form of a list 
  print(aut.perm)     
  # Show the automorphism 'aut' with its permutation number 
  print(aut)     

**Remark**

Class |AutPl| is a subclass of class 
``mmgroup.structures.AbstractGroupWord`` which implements
elements of an abstract group and the operations on such elements.
For an instance ``g`` of |AutPl|, the object ``g.group`` is
an instance of a subclass of  ``mmgroup.structures.AbstractGroup``.
That subclass models the group
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`.

"""

import re

from functools import reduce
from operator import __or__
from numbers import Integral, Number
from random import randint

import numpy as np

from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.abstract_group import AbstractGroup
from mmgroup.structures.parse_atoms import AtomDict, ihex     
from mmgroup.structures.parse_atoms import eval_atom_expression        



try:
    # Try importing the fast C function
    from mmgroup import mat24 
    from mmgroup.mat24 import MAT24_ORDER
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24
    MAT24_ORDER = Mat24.MAT24_ORDER
    





ERR_TYPE = "unsupported operand types for %s: '%s' and '%s'"
ERR_RAND = "Illegal string for constricting type %s element" 



#######################################################################
# Import derived classed
#######################################################################

import_pending = True
Cocode = None


def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending, Cocode
    from mmgroup.structures.cocode import Cocode
    import_pending = False




#######################################################################
# Function implementing the constructor for class AutPL
#######################################################################



ERR_AUTPL_P1 = "AutPL expects at most 1 parameter of type %s" 
ERR_AUTPL_TYPE = "Cannot construct AutPL instance from type %s" 


autpl_conversions = {
}

def autpl_from_obj(d = 0, p = 0, unique = 1):
    """Try to convert tuple (d, p) to a Parker loop automorphism.

    Parameter ``d`` decribes a element of the Golay cocode as in the
    constructor of class ``Cocode``. It may be of type ``int``, ``str``
    or an instance of class ``Cocode``. Pt defaults to ``0``.

    Alternatively, ``d`` may be an instance of class ``AutPL``. In this 
    case, ``p`` must be set to its default value.


    Parameter ``p`` describes a element of the Mathieu group ``Mat24``.
    It defaults to the neutral element of ``Mat24``. It may be
    
     * An integer encoding the number of a permutation in ``Mat24``.

     * A list encoding a permutation in the Mathieu group ``Mat24``.

     * A zip object or a dictionary encodes such a permutation as
       a mapping. That mapping must contain sufficiently many 
       entries to be unique.  

     * The string 'r' encodes a random permutation in ``Mat24``.

    If parameter ``unique`` is ``True`` (default), then the parameter
    ``p`` must either be ``r`` or describe a unique permutation in
    ``Mat24``. Otherwise lowest possible permutation (in 
    lexicographic order) is taken.

    The function returns a pair ``(cocode, permutation_number)``.
    """
    if import_pending:
        complete_import()
    if isinstance(d, Integral):
        d_out = int(d)
    elif isinstance(d, Cocode):
        d_out = d.value
    elif isinstance(d, AutPL):
        d_out, p_out =  d._cocode, d._perm_num
        if p:
            raise TypeError(ERR_AUTPL_P1 % type(d))
        return d_out, p_out
    elif isinstance(d, str):
        d_out = Cocode(d).value
    else: 
        try:
            f = autpl_conversions[type(d)]
        except KeyError:
            raise TypeError(ERR_AUTPL_TYPE % type(d))
        if p:
            raise TypeError(ERR_AUTPL_P1 % type(d))
        d_out, p_out =  f(d)
        return d_out, p_out
            
    if p == 'r':
        return d_out, randint(0, MAT24_ORDER - 1)
    if isinstance(p, Integral):
        if 0 <= p <  MAT24_ORDER:
            return d_out, int(p)
        err = "Bad number for permutation in Mathieu group  Mat24"
        raise ValueError(err) 
    if isinstance(p, (list, range)):
        if mat24.perm_check(p):
            err = "Permutation is not in Mathieu group M_24"
            raise ValueError(err)
        return d_out, mat24.perm_to_m24num(p)
    if isinstance(p, zip):
        p = dict(p)
    if isinstance(p, dict):
        h1, h2 = [list(t) for t in zip(*p.items())]
        res, perm = mat24.perm_from_map(h1, h2)
        if res == 1 or (res > 1 and not unique):
            return d_out, mat24.perm_to_m24num(perm) 
        if res < 1:
            err = "Permutation is not in Mathieu group M_24"
        else:                 
            err = "Permutation in Mathieu group M_24 is not unique"
        raise ValueError(err)






#######################################################################
# Class AutPL
#######################################################################


            



class AutPL(AbstractGroupWord):
    r"""This class models a standard automorphism of the Parker loop.

    :param d:

      This parameter describes an element of the Golay cocode. If ``d``
      is an instance of class |AutPL| then this describes an
      automporphism of the Parker loop; in that case parameter ``p``
      must be set to its default value. Legal types of parameter ``d`` 
      see  below. The parameter defaults to zero word of the cocode.
       
    :param p:

      This parameter describes a permutation in the Mathieu group 
      ``M_24`` or, more precisely, the standard representative of 
      that permutation in the automporphism group of the Parker 
      loop. Legal types of that parameter see below.
      The parameter defaults to neutral element of ``Mat24``.

    :param unique:

       If this is ``True`` (default) and parameter ``p`` is not ``r``
       then parameter ``p`` must describe a unique permutation in
       the Mathieu group ``M_24``. Otherwise we take the least 
       possible permutation (in lexicographic order) that agrees
       with  parameter ``p``.
 
    :return: A standard Parker loop automorphism
    :rtype:  an instance of class |AutPL|

    :raise:
        * TypeError if ``type(value)`` is not in the table given below.
        * ValueError if the set of arguments cannot be converted to an
          instance of class  |AutPL|.


    .. table:: Legal types for parameter ``d`` in constructor of class ``AutPL``
      :widths: 25 75

      ===================== ==================================================
      type                  Evaluates to
      ===================== ==================================================
      ``int``               The number of an element of the Golay cocode.

      class |Cocode|        This represents an element of the Golay cocode.

      ``str``               Create random element depending on ``str``
                              | ``'r'``: Create an arbitrary cocode element
                              | ``'e'``: Create an even cocode element
                              | ``'o'``: Create an odd cocode element

                            Any other string is illegal.

      class |AutPL|         A deep copy of the given automorphism 
                            in |AutPL| is returned. Then parmeter ``p``
                            must be set to its default value.

      ===================== ==================================================


    .. table:: Legal types for parameter ``p`` in constructor of class ``AutPL``
      :widths: 25 75

      ===================== ==================================================
      type                  Evaluates to
      ===================== ==================================================
      ``int``               Here the integer is the number of a permutation  
                            in the Mathieu group ``M_24``. 

      ``list`` of ``int``   A list ``l_p`` of ``24`` of disjoint integers
                            ``0 <= i < 24`` is interpreted as a permutation
                            in ``M_24`` that maps ``i`` to ``l_p[i]``.

      ``dict``              A dictionary specifies a mapping from a subset
                            of the integers ``0 <= i < 24`` to integers
                            ``0 <= dict[i] < 24``. This mapping must
                            extend to a permutation in ``M_24``. 
                            If parameter  ``unique`` is ``True`` (default)
                            then that permutation must be unique.

      ``zip`` object        ``zip(x,y)`` is equivalent to 
                            ``dict(zip(x,y))``.

      ``str``               Create random element depending on ``str``
                              | ``'r'``: Create an random element of ``M_24``
                            
                            Any other string is illegal.
      ===================== ==================================================



    Let ``a`` be an instance of class |GcVector|, |GCode|, |Cocode|,
    or |PLoop|, and let ``g1`` , ``g2`` be instances of class |AutPL|. 
    Then  ``a * g1`` is the result of the natural operation of ``g1`` 
    on  ``a``, which belongs to the same class as ``a``.

    ``g1 * g2``  means group multiplication, and ``g1 ** n`` means
    exponentiation of ``g1`` with the integer ``n``. ``g1 ** (-1)`` 
    is the inverse of ``g``. ``g1 / g2`` means ``g1 * g2 ** (-1)``.

    ``g1 ** g2`` means ``g2**(-1) * g1 * g2``.    
    """
    __slots__ = "_cocode", "_perm_num", "_perm", "rep"
    ERR_UNDEF = "Parker loop automorphism is not defined by input data"
    _perm_ = list(range(24))  # neutral permutation
    _rep_ = mat24.perm_to_autpl(0, _perm_)
    group = None       # will be set to StdAutPlGroup later
    
    def __init__(self, d = 0, p = 0, unique = 1):
        self._cocode, self._perm_num = autpl_from_obj(d, p, unique)
        self._compute_from_numbers()

    def _compute_from_numbers(self):
        self._perm = mat24.m24num_to_perm(self._perm_num)
        self.rep = mat24.perm_to_autpl(self._cocode, self._perm)
            
    def _compute_from_rep(self):
        self._perm =  mat24.autpl_to_perm(self.rep)
        self._perm_num = mat24.perm_to_m24num(self._perm)
        self._cocode = mat24.autpl_to_cocode(self.rep)

    def check(self):
        """Check automorphism for consistency via 'assert' statements

        ``a.check()`` returns ``a``.
        """
        assert self._perm ==  mat24.autpl_to_perm(self.rep)
        assert self._perm_num == mat24.perm_to_m24num(self._perm)
        assert 0 <= self._perm_num < mat24.MAT24_ORDER
        assert self._cocode == mat24.autpl_to_cocode(self.rep)
        return self

    @property    
    def parity(self):
        """Return parity of an automorphism
 
        This is ``s`` if ``PLoopOmega`` maps to ``(-1)**s * PLoopOmega``
        for ``s == 0`` or ``s == 1``.
        """
        return (self._cocode >> 11) & 1

    @property    
    def cocode(self):
        """Return number of cocode element in the automorphism.

        For an instance ``g`` of |AutPL| we have

          ``g == AutPL(Cocode(g.cocode)) * AutPL(g.perm)``.
        """
        return self._cocode

    @property        
    def perm_num(self):
        """Return the number of the permutation of the automorphism

        The Mathieu group ``M_24`` has ``244823040`` elements 
        numbered from ``0`` to ``244823039`` in lexicographic
        order. Thus the neutral element has number ``0``. 
        """
        return self._perm_num

    @property    
    def perm(self):
        """Return permutation of automorphism as a list ``p_l``.

        Then the permutation maps ``i`` to ``p_l[i]`` for
        ``i = 0,...,23``.
        """
        return self._perm

    @property    
    def mmdata(self):
        return np.array([
            0x10000000 + (self._cocode & 0xfff),
            0x20000000 + self._perm_num
        ], dtype = np.uint32)

    def as_tuples(self):
        return [('d', self._cocode), ('p', self._perm_num)]
       



#######################################################################
# Class AutPlGroup
#######################################################################


def autpl_element_from_obj(g, t):
    res =  AutPL()
    res._cocode, res._perm_num = autpl_from_obj(t)
    res._compute_from_numbers()
    return res

@singleton
class AutPlGroup(AbstractGroup):
    word_type = AutPL              # type of an element (=word) in the group
    is_mmgroup = True
    #atom_parser = {}               # see method parse()
    #rand_masks  = {"r":(0xfff,0), "e":(0x7ff,0), "o":(0x7ff,0x800)} 
    conversions = {
      #  list: autpl_element_from_obj,
      #  zip: autpl_element_from_obj,
      #  dict: autpl_element_from_obj,
      #  AutPL: autpl_element_from_obj,
    }
    #FRAME =  AUTPL_FRAME
    ERR_PERM_NUM = "Illegal permutation number for Mathieu group M_24"

    def __init__(self):
        super(AutPlGroup, self).__init__()
        self.atom_parser = AtomDict(self.atom)


    def __call__(*args, **kwds):
        err = "Class AutPlGroup object is not callable"
        raise TypeError(err)

    def atom(self, tag = None, data = None):
        err = "Class AutPlGroup has no attribute 'atom'"
        raise AttributeError(err)

    def _imul(self, g1, g2):
        g1.rep = mat24.mul_autpl(g1.rep, g2.rep)
        g1._compute_from_rep()
        return g1    

    def _invert(self, g1):
        res = AutPL()
        res.rep = mat24.inv_autpl(g1.rep)
        res._compute_from_rep()
        return res

    def copy_word(self, g1):
        res = AutPL()
        res._cocode, res._perm_num = g1._cocode, g1._perm_num
        res._perm, res.rep =  g1._perm[:], g1.rep[:]
        return res

    def _equal_words(self, g1, g2):
        return g1._cocode == g2._cocode and g1._perm_num == g2._perm_num

        

    def str_word(self, g):
        """Convert group atom g to a string

        """
        s = []
        if g._cocode:
            s.append("d_" + ihex(g._cocode))
        if g._perm_num:
            s.append("p_%d" % g._perm_num)
        s = "*".join(s) if len(s) else "d_0"
        return "AutPL<%s>" % s



StdAutPlGroup = AutPlGroup()   # This is the only instance of AutPlGroup

AutPL.group = StdAutPlGroup





from mmgroup.generate_c.generate_functions import prepend_blanks
from mmgroup.generate_c.generate_functions import format_item
from mmgroup.generate_c.generate_functions import UserDirective
from mmgroup.generate_c.generate_functions import UserFormat
from mmgroup.generate_c.generate_functions import make_table

from mmgroup.generate_c.make_c_tables import TableGenerator
from mmgroup.generate_c.make_c_tables import make_doc
from mmgroup.generate_c.make_c_tables import c_snippet

from mmgroup.generate_c.make_pyx import pxd_to_pyx
from mmgroup.generate_c.make_pyx import pxd_to_function_list






"""An exprimental free group with relations


""" 
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import re
from random import randint

from mmgroup.structures.parse_atoms import TaggedAtom
from mmgroup.tests.groups.auto_group import AutoGroup
from mmgroup.tests.groups.auto_group import AutoGroupMulError

####################################################################
####################################################################
### An experimental free group with relations for testing
####################################################################
####################################################################


def f_const(c):
    """Return a function that returns list of tags c on any input"""
    def f(*args):
        return [TaggedAtom(b) for b in c]
    return f

def _failure(*args):
    raise AutoGroupMulError




class FreeGroup(AutoGroup): 

    def __init__(self, generators = "a", commute="", rewrite = {}, bad_commute=""):
        """Model a free group with n generators

        Generators are named a,b,c..., their inverses are A,B,C... .

        If commute is a string then all characters in that string
        commute.
        rewrite is optional dictionary with keys and values both
        being strings. Then the items of the dictionary are
        interpreted as rewriting rules. 

        bad_commute gives an experimental bad matching rule for 
        making all characters in bad_commute commuting. Here every
        commuting pair is matched, and the substitution function
        raises AutoGroupMulError if the pair should should not be
        echanged.
        """
        rules = {}        
       
        for arg in [generators, commute, bad_commute]:
            assert arg == "" or arg.islower()
        for g in generators:
            G = g.upper()
            rules[G + g] = rules[g + G] = f_const("")
           
        for c1 in commute + commute.upper():
            c1l = c1.lower()
            for c2 in commute + commute.upper():
                c2l = c2.lower()
                if c1 != c2:
                    if c1l > c2l:
                        rules[c1 + c2] = f_const(c2 + c1)
                    elif c1 in bad_commute and c2 in bad_commute: 
                        rules[c1 + c2] = _failure
 
        super(FreeGroup, self).__init__({}, rules, {})

    def equal_atoms(self, atom1, atom2):
        return atom1.tag == atom2.tag


    def reduce_atom(self, atom):
        return atom

    def iter_inv(self, atom):
        t = atom.tag   
        yield TaggedAtom(t.lower() if t.isupper() else t.upper())

             
    def __call__(self, w=""):
        """Return the word of the free group given by the string w"""
        assert isinstance(w, str)
        res = self.word_type(*[TaggedAtom(c) for c in w], group = self)
        return res.reduce()

    def str_word(self, word):
        x = "".join((str(atom.tag) for atom in word.iter_generators()))
        return x if x else "<1>"
		




"""Zip the whole project to a zip file in subdirectory 'build'.


The zip process is based on:

python setup.py sdist --format=zip

Non-source files are removed from the zip file.

"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import zipfile, time
import sys, os
import subprocess

from zipfile import ZIP_DEFLATED


tm = time.strftime("%Y_%m_%d_%H_%M")

filename = r'Monster_%s.zip' % tm


source = r"dist\mmgroup-0.0.7.zip"

zip_path = os.path.join('backup', filename)

backup_dir = r"E:\AlterPC\projects\MonsterSup\backup"




def is_to_be_zipped(filename):
    if filename[-4:] in [".pyd", ".dll"]:
        return False
    if filename.find("build/") >= 0:
        return False 
    if filename.find("docs/build/") >= 0:
        return False 
    if filename.find("egg-info") >= 0:
        return False 
    if filename.find("pytest_cache") >= 0:
        return False 
    if filename.find("/src/mmgroup/dev/c_files/") >= 0:
        return False 
    if filename.find("/src/mmgroup/dev/c_doc/") >= 0:
        return False 
    if filename.find("/src/mmgroup/dev/pxd_files/") >= 0:
        return False        
    return True


if __name__ == '__main__':
    if os.path.exists(source):
        os.remove(source)

    subprocess.check_output([sys.executable, "setup.py", "sdist", "--format=zip"]) 

    zin = zipfile.ZipFile (source, 'r')
    zout = zipfile.ZipFile (zip_path, 'w', compression=ZIP_DEFLATED,  compresslevel=9)
    for item in zin.infolist():
        buffer = zin.read(item.filename)
        #print(item.filename, is_to_be_zipped(item.filename))
        if is_to_be_zipped(item.filename):
            zout.writestr(item, buffer)
    zout.close()
    zin.close()

    try:
        print( subprocess.check_output(["cmd", "/c", "copy", zip_path, backup_dir]) )
        zf = zipfile.ZipFile(os.path.join(backup_dir, os.path.join(backup_dir,filename)))
        print ("Backup to external drive successful!")
    except:
        print ("Backup to external drive did not work!")
        raise




from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import pytest

import types
import sys
import re
import os
from operator import __or__, __xor__, __and__
from functools import reduce
import random


 
from mmgroup.bitfunctions import v2, lmap, lrange

from mmgroup.generate_c  import prepend_blanks
from mmgroup.dev.mat24.mat24aux import generate_golay24_decode


def lsbit24(x):
    return v2(x | 0x1000000)


from mmgroup.dev.mat24.mat24heptad import HeptadCompleter
from mmgroup import mat24
from mmgroup.dev.mat24.mat24_ref import Mat24


#########################################################################
# Test completion of heptads and octads
#########################################################################


def random_umbral_heptad(gc, as_heptad_at_pos = None):
    """Return random list p with p[0],...,p[5], p[8] an umbral heptad.

    That list hat length 24 (with unused entries set to None) and 
    it can be completed to a unique permutation in Mat24.

    So p[0],...,p[5]  are contained in an octad. The special element
    p[8] is not contained in that octad.

    if 0 <= as_heptad_at_pos < 7 is given then the heptad is returned
    as a list of 7 elements with the special element at that position.

    Parameter gc must be an instance of class Mat24.     
    """
    p = []
    while len(p) < 5:
        x = random.randint(0,23)
        if not x in p:
            p.append(x)
    as_cocode = gc.vect_to_vintern(sum(1 << x for x in p))
    syn = gc.syndrome_table[ as_cocode & 0x7ff ]
    syn = [ (syn >> i) & 31 for i in (0, 5, 10) ]
    octad, p8 = p + syn, None
    while p8 is None:
        x = random.randint(0,23) 
        if not x in octad:
            p8 = x
    p.append(syn[random.randint(0,2)])
    if not as_heptad_at_pos is None:
        return p[:as_heptad_at_pos] + [p8] + p[as_heptad_at_pos:]         
    return p + [None, None, p8] + [None]*15
     
   


def get_testdata(gc):
    hc = gc.heptad_completer
    import random
    prod = lrange(24)
    yield prod, True
    for i in range(100):
        p = random_umbral_heptad(gc)
        yield p, True
        hc.complete_heptad(p) 
        prod =  gc.mul_perm(prod, p)
        #print( prod )
        yield p, True

    # test some error cases
    for i in range(100):
        p = random_umbral_heptad(gc)
        hc.complete_heptad(p)
        p6 = p[6]
        j = random.randint(0,5)
        p[j], p[8] = p[8], p[j]
        yield p,  False
        p[j], p[8] = p[8], p[j]
        yield p,  True
        ok_j = p[j]
        p[j] = p[8]
        yield p,  False
        p[j] = random.randint(24,32)
        yield p,  False
        p[j] = ok_j
        yield p,  True
        p[8] = p6
        yield p,  False
        p[8] = p[j]
        yield p, False
        p[8] = random.randint(24,32)
        yield p,  False
    for i in range(6):
        for j in range(i):
            p = random_umbral_heptad(gc)
            hc.complete_heptad(p)
            p[i] = p[j]
            yield p, False     
              
        
  
def heptad_testdata(gc):
    hc = gc.heptad_completer
    import random    
    pos = random.randint(0,6)    
    for i in range(50):
        h1 = random_umbral_heptad(gc, pos)
        h2 = random_umbral_heptad(gc, pos)
        yield h1, h2


@pytest.mark.mat24
@pytest.mark.parametrize("gc, gc_ref", [(mat24, Mat24)])
def test_heptad_completer(gc, gc_ref):
    hc = gc_ref.heptad_completer
    for p, ok in get_testdata(gc_ref):
        p_in = p[:6] + [None]*2 + p[8:9] + [None]*15
        if ok:
            err =  gc.perm_complete_heptad(p_in)
            #assert  err == 0, (p_in, hex(err))
            if not None in p:
                assert list(p) == list(p_in), (p, p_in, hex(err))
                i = hc.perm_to_int(p)
                assert 0 <= i < 244823040
                p1 = hc.int_to_perm(i)
                assert list(p) == list(p1), (i, p, p1)
        else:
            with pytest.raises(ValueError):
                gc.perm_complete_heptad(p_in)
        if ok:
            hexad = p[:6] + [None]*2
            hexad_ref = hexad[:]
            gc.perm_complete_octad(hexad)
            assert hexad == p_in[:8]
            gc_ref.perm_complete_octad(hexad_ref)
            assert hexad_ref == p_in[:8]

    for h1, h2 in heptad_testdata(gc_ref):
        p = gc.perm_from_heptads(h1, h2)
        for i in range(7):
            assert p[h1[i]] == h2[i]
    perm0 = gc.perm_to_m24num(lrange(24))
    assert perm0 == 0, perm0

    print("Heptad completer test passed")



#########################################################################
# Test completion of heptads and dodecads
#########################################################################

def random_dodecad():
    while True:
        gc = random.randint(0,0xfff)
        if mat24.gcode_weight(gc) == 3:
            v = mat24.gcode_to_vect(gc)
            w, l = mat24.vect_to_bit_list(v)
            assert w == 12
            l = l[:12]
            random.shuffle(l)
            return l


def get_test_dodecads():
    for i in range(200):
        yield random_dodecad(), random_dodecad()


@pytest.mark.mat24
def test_perm_from_dodecad():
    for ntest, (d1, d2) in enumerate(get_test_dodecads()):
        #print(d1, d2)
        perm = mat24.perm_from_dodecads(d1[:9], d2[:9])
        for i in range(5):
             assert perm[d1[i]] == d2[i]
        img = set([perm[x] for x in d1])
        assert img == set(d2)

    

#########################################################################
# Test function  mat24_perm_from_map()
#########################################################################



def rand_perm():
    """Return random permutation in ``Mat24`` as a list"""
    num = random.randint(0, mat24.MAT24_ORDER-1)
    return mat24.m24num_to_perm(num)


def map_vect(v):
    """Apply random permutaiton in in ``Mat24`` of vector

    Here ``v`` is a vector of integers ``0 <= v[i] < 24``.
    The function generates a random permutation ``p`` in
    the group ``Mat24`` and returns the list
    
    [p[0], p[1],...,p[22], p[23]].
    """
    pi = rand_perm()
    return [pi[x] for x in v]


def perm_from_map_testdata(ntests=10):
    """Yield test data for testing function perm_from_map()

    The function yields quadruples (h1, h2, ref_res, ref_p).

    The function perm_from_map() is expected to find a
    permutation p that maps the entries in the list h1 to
    the entries in the list h2, provided that ``ref_res``
    is greater than zero.

    Also, function perm_from_map() should return the value
    ``ref_res``. If ``ref_p`` is not None then the permutation
    returnd by function perm_from_map() should be equal
    to ``ref_p``.    
    """
    Id = list(range(24))
    yield list(range(5)), list(range(5)), 3, Id
    yield [], [], 3, Id
    yield list(range(9)), list(range(9)), 1, Id
    yield [0,1,2,3,4,5,6], [0,1,2,3,4,5,8], 0, None

    PERMS = [
         ([0,1,2,3,4,5], 3),
         ([0,1,2,3,4,9], 2),
         ([0,1,2,3,4,5,9], 1),
         ([0,1,2,3,4,5,6,9], 1),
    ]
    for i in range(1, 9):
        PERMS.append( (range(i), 3) )
    for perm, res in PERMS:
        yield perm, perm, res, Id 
        continue       
        v = map_vect(perm)
        yield v, v, res, Id
        for i in range(ntests):
            v1, v2 = map_vect(perm), map_vect(perm)
            yield v1, v2, res, None
            
    for n in range(9, 25):
        L = random.sample(range(24),n)
        yield L, L, 1, Id
        for j in range(ntests):
            p = rand_perm()
            ind = random.sample(range(24), n)
            perm = [p[i] for i in ind]
            yield ind, perm, 1, p
    

def one_test_perm_from_map(h1, h2, ref_res, ref_p,verbose = 0):
    if verbose:
        print("h1 =", h1)
        print("h2 =", h2)
        print("Expected return value:", ref_res)
    res, perm = mat24.perm_from_map(h1, h2)
    ok, ok_perm = res == ref_res, True
    if ref_p:
        ok_perm = perm == ref_p
    elif ref_res > 0:
        for i, x in enumerate(h1):
            ok_perm = ok_perm and perm[x] == h2[i]
    ok = ok and ok_perm
    if not verbose and not ok:
        print("h1 =", h1)
        print("h2 =", h2)
        print("Expected return value:", ref_res)
    if verbose or not ok:
        print("Obtained return value:", res)
        print("p =", perm)
        if res <= 0:
            if ref_p and ref_p != perm:
                print("Expected:\n   ", ref_p)
        if not ok_perm:
            print("Permutation p is bad")
        if res != ref_res:
            print("Return value is bad")
    if not ok:
        err = "Test of function mat24.perm_from_map failed"
        raise ValueError(err)

    if ref_p or len(h1) <= 1:
        return

    # If no expected permutation is given, shuffle h1 and h2
    # in the same way and test once more with shuffled h1, h2.
    lh = len(h1)
    ind = list(range(lh)) if lh > 2 else [1,0]
    if (lh > 2):random.shuffle(ind)
    h1a = [h1[i] for i in ind]
    h2a = [h2[i] for i in ind]
    res1, perm1 = mat24.perm_from_map(h1a, h2a)
    assert res1 == res, (res1, res)
    assert perm1 == perm, (perm1, perm)


@pytest.mark.mat24
def test_perm_from_map(verbose = 0):
    for n, (h1, h2, ref_res, ref_p) in enumerate(perm_from_map_testdata()):
        if verbose: print("\nTest", n+1)
        one_test_perm_from_map(h1, h2, ref_res, ref_p, verbose) 



#########################################################################
# Test lexical peroperty of function  mat24_perm_from_map()
#########################################################################

 

def rand_perm_vect(v, v_more):
    """

    Let ``v`` be a list representing a vector in ``GF(2)^24``
    and ``v_more`` be a list such that ``v + v_more`` is or 
    contains an umbral heptad.

    The function returns to random images ``h1, h2`` of ``v`` 
    under ``Mat24``, and also a (somehow) random permutation
    ``pi_ref`` that maps ``h1`` to ``h2``.
    """
    
    pi1 = rand_perm()
    pi2 = rand_perm()
    h1 = [pi1[x] for x in v]
    h2 = [pi2[x] for x in v]
    vm = v + v_more
    h1m = [pi1[x] for x in vm]
    h2m = [pi2[x] for x in vm]
    return h1, h2, mat24.perm_from_map(h1m, h2m)[1]
    

def perm_map_lex_testdata(ntests = 100):
    yield [], [], list(range(24)), 3
    STD_V = [0,1,2,3,4,5,6,7,8]
    for i in range(ntests):
        for j in range(6, 9):
            h1, h2, pi_ref = rand_perm_vect(STD_V[:j], STD_V[j:])
            yield h1, h2, pi_ref, 3
        UMBRAL_HEXAD = [0,1,2,3,4,8] 
        MORE = [9]
        h1, h2, pi_ref = rand_perm_vect(UMBRAL_HEXAD, MORE)
        yield h1, h2, pi_ref, 2
   
def one_test_perm_map_lex(h1, h2, pi_ref, ref_res = None, verbose = 1):
    errors = 0
    if verbose:
        print("h1 =", h1)
        print("h2 =", h2)
        print("pi_ref:", pi_ref)
        print("Expected return value:", ref_res)
    assert [pi_ref[x] for x in h1] == h2
    res, pi = mat24.perm_from_map(h1, h2)
    if verbose:
        print("pi:", pi)
        if pi > pi_ref:
            print("pi is greater than pi_ref")
            errors += 1
        if res != ref_res:
            print("Obtained return value:", res)
        #assert pi <= pi_ref
    assert [pi[x] for x in h1] == h2
    if errors > 0:
        print(errors, "lexical errors")
    assert res == ref_res
    assert errors == 0
     
@pytest.mark.mat24
def test_perm_map_lex(verbose = 0):
    for n, (h1, h2, pi_ref, ref_res) in enumerate(perm_map_lex_testdata()):
        if verbose: print("\nTest", n+1)
        one_test_perm_map_lex(h1, h2, pi_ref, ref_res, verbose) 

    


r"""

The functions in module ``gen_xi_functions.c`` deal with the 
operation of the generator :math:`\xi` of the monster defined 
in :cite:`Seysen20`, section 9, on a subgroup :math:`Q_{x0}` and
also with the monomial part of that operation on the  
196884-dimensional representation :math:`\rho` of the monster.

Generator :math:`\xi` is an element of the subgroup 
:math:`G_{x0}` of the monster. :math:`G_{x0}` operates on its 
normal subgroup :math:`Q_{x0}` of  of structure :math:`2^{1+24}`
by conjugation as described in the **Guide**, section 
:ref:`computation-leech2`. :math:`Q_{x0}` is extraspecial 
and has center :math:`\{\pm1\}`. We have 
:math:`Q_{x0} / \{\pm 1\} \cong \Lambda / 2 \Lambda`. An element 
:math:`x` of :math:`Q_{x0}` is short if it corresponds to
a vector in the Leech lattice :math:`\Lambda` of norm 4. 
:math:`G_{x0}` also operates on the :math:`2 \cdot 98280` short 
elements of :math:`Q_{x0}`.

Representation :math:`\rho` has a subspace :math:`98280_x` of 
dimension 98280 on which :math:`G_{x0}` operates monomially in the 
same way as it operates on the short elements of :math:`Q_{x0}`. 
So the basis vectors of  :math:`98280_x` (together with their opposite 
vectors) can be identified with the short elements of :math:`Q_{x0}`.

The functions in module ``gen_xi_functions.c`` are also used for 
computing tables describing the monomial operation of 
:math:`\xi^e, e=1,2` on :math:`98280_x`.

Module ``mmgroup.dev.generators.gen_xi_ref`` is a pure python 
substitute for this set of C functions; but calculating the tables
with python takes a rather long time.

In  section :ref:`mmrep-label` we use the following names and tags 
for the basis vectors of :math:`98280_x`.


    ================ === =================================== ========
    Name             Tag Entries                             Remarks
    ================ === =================================== ========
    :math:`X^+_{ij}`  B  i, j;  0 <= j < i < 24              (1)
    :math:`X_{ij}`    C  i, j;  0 <= j < i < 24              (1)
    :math:`X_{o,s}`   T  o, s;  0 <= o < 759, 0 <= s < 64    (2)
    :math:`X_{d,j}`   X  d, j;  0 <= d < 2**11, 0 <= j < 24  (1),(3)
    ================ === =================================== ========

Remarks

(1)  i and j, 0 <= i,j < 24  refer to basis vectors of the
     Boolean vector space in which the Golay code is defined.
(2)  o is one of 759 octads, s is one of 64 even subsets of
     octad d (modulo its complement in d), as described in
     section :ref:`octads_label`.
(3)  d with 0 <= d < 2048 refers to the Parker loop element d. 
     For larger values d we have:
     
     .. math::
     
        X_{2048+k,j} = X_{k,j}, \; Y_{2048+k,j} = -Y_{k,j},  \;
        Z_{2048+k,j} = Z_{k,j},
        
        U_{4096+k,j} = -U_{k,j},  \quad  \mbox{for} \;  U = X, Y, Z;
        \; 0 <= k < 2048 \, .

We group these basis vectrors into 5 boxes (labelled 1,...,5) with 
each box containing at most :math:`3 \cdot 2^{13}` entries. Element 
:math:`\xi` permutes these boxes as follows:

   Box1 -> Box1,  Box2 -> Box2,  Box3 -> Box4 -> Box5 -> Box3 .  (1)

The mapping from the basis vectors to entries in boxes is:

   ===================== ===     =======================
   Basis vector          Box     Entry
   ===================== ===     =======================
   B[i,j]                 1         0 +  32 * i + j
   C[i,j]                 1       768 +  32 * i + j
   T[o,s], o < 15         1      1536 +  64 * o + j             
   T[o,s], 15 <= o < 375  2        64 * (o - 15) + j           
   T[o,s], o >=  375      3        64 * (o - 375) + j
   X[d,j], d <  1024      4        32 *  d + j
   X[d,j], d >= 1024      5        32 * (d - 1024) + j
   ===================== ===     =======================

This subdivision looks weird, but is has quite a few advantages:

  * The lower index (j or s) has stride 1, and the the stride of the
    higher index (i, o or d) is the lowest possible power of two. So 
    accessing an entry is easy. In the C code for a representation
    derived from :math:`\rho` the entries of a vector of such a 
    representation are strided in the same way. So the tables 
    computed in this module can be used in the C code for operator
    :math:`\xi`.

  * An entry in a box is always less than :math:`2^{15}`, so any 
    entry can be stored in a 16-bit integer, together with a sign bit.  

  * Boxes are permuted as above, so 4 tables of 16-bit integers
    and size <=  :math:`3 \cdot 2^{15}` are sufficent to encode 
    the operation of :math:`\xi`.

We remark that boxes 3 and 5 contain the short vectors with odd
Golay code words; the other boxes contain those with even
code words. Here the parity (even/odd) of a Golay code word means 
the scalar product with the cocode word :math:`\omega` which 
has six 1-bits  in column 0 of the MOG, see :cite:`Seysen20`, 
section 2.2. Short vectors with an odd cocode element occur 
precisely in boxes 4 and 5.

A similar, but finer  subdivision of the whole space :math:`\rho`
(and not only of the subspace :math:`98280_x`) is given in 
:cite:`Iva09`, section 3.4.


The operation of the generator :math:`\xi` on the group :math:`Q_{x0}`
......................................................................

Using :cite:`Seysen20`, Lemma 9.5, we can easily compute the
values :math:`\xi^{-e} x \xi^e` for all short :math:`x \in Q_{x0}`
and :math:`e = 1,2`. Here we represent an  :math:`x \in Q_{x0}`
as a 25-bit integer in Leech lattice encoding as described above.

Let :math:`\mathcal{C}` and :math:`\mathcal{C}^*` be the Golay 
code and its cocode. In :cite:`Seysen20`, section 2.2 we give 
decompositions
:math:`\mathcal{C} = \mathcal{G} \oplus \mathcal{H}` and
:math:`\mathcal{C}^* = \mathcal{G}^* \oplus \mathcal{H}^*`.

The formulas in ibid., Lemma 9.5 use auxiliary functions 
:math:`\gamma : \mathcal{G} \rightarrow \mathcal{G^*}` and
:math:`w_2 : \mathcal{G} \cup \mathcal{G^*} \rightarrow
\mathbb{F}_2` defined in ibid., section 3.3. Using the
decompositions above, we may extend :math:`\gamma` to
:math:`\mathcal{C}` and :math:`w_2` to
:math:`\mathcal{C} \cup \mathcal{C^*}`.



Short vector encoding of the short vectors in :math:`Q_{x0}`
.............................................................
  
For each short vector in :math:`Q_{x0}` we compute the number of 
the box and the entry in the box using the information given in (2). 
Then we store the vector in the lowest 19 bits of a 32-bit integer 
as follows:

  *  Bit 18...16:  number of the box
  
  *  Bit 15:       sign bit 
  
  *  Bit 14...0:   Entry in the box

Tables for the operation of :math:`\xi` and :math:`\xi^2` contain the 
lower 16 bits of that encoding. Bits 18...16 can be reconstructed 
from the permutation of the boxes given by (1).

We precomute a set of tables containing  :math:`2 \cdot 98260`
values :math:`\xi^{-e} x \xi^e`, where the short element 
:math:`x` of :math:`Q_{x0}`  is given in 
**Short vector encoding**.


"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from mmgroup.bitfunctions import bw24, lmap

import random
import numpy
from numpy import array, zeros, uint8, uint16, uint32

from mmgroup.dev.mat24.mat24_ref import Mat24


#######################################################################
# Auxiliary functions 
#######################################################################


def w2_gamma(v):
    """Implement function gamma() in [Seysen20], section 3.3.

    Given a Golay code vector v  in 'gcode' representation, see
    module mat24.py, we return a pair w2, c. Here c is equal to
    the element gamma(v) of the Golay cocode, with gamma() as in 
    [Seysen20], section 3.3. ,c is returned in 'cocode' representation, 
    see module mat24.py. We also return the bit w2 = w2(c), with w2()
    as defined in [Seysen20], section 3.3.   
    """
    x = Mat24.gcode_to_vect(v)
    x1 = sum( (x >> i) & 0x111111 for i in range(1,4) )
    x1 = (x1 >> 1) & 0x111111
    w2 = bw24(x1)
    w2 = ((w2 * (w2 - 1)) >> 1) & 1
    c = Mat24.vect_to_cocode(x1)
    return w2, c



def compress_gray(x):
    """Extract the gray part of a Colay code or cocode word

    The function returns the 'gray' part of a Golay code or cocode
    as a 6-bit number. It drops the 'colored' part of the word.

    A 12-bit Golay code or cocode word can be expressed as a sum of
    a 6-bit 'gray' and a 6-bit 'colored' word, see [Seysen20], 
    section 2.2.  

    A Golay code word x must be given in 'gcode' and a cocode word 
    must be given in 'cocode' representation. 
    """
    return (x & 0x0f) + ((x >> 6) & 0x30)

def expand_gray(x):
    """Reverse the effect of function compress_gray()

    Given a compressed 'gray' Golay code or cocode word, the function 
    returns the original 'gray' word in code or cocode representation.
    """
    return (x & 0x0f) + ((x & 0x30) << 6) 



#######################################################################
# Class Mat24Xi
#######################################################################



class GenXi(object):
    """Operation of xi of a group G_x0 on the normal subgroup Q_x0.

    A subgroup G_x0 of shape 2**{1+24}.Co_1 of the Monster group is 
    used for the construction of the monster group M in [Con85]. G_x0 
    is generated by a subgroup N_x0 of shape 2**{1+24}.2**{11}.Mat24, 
    and an additional element xi of order 3, see [Seysen20], section 9. 
    Mat24 is the Mathieu group acting on 24 elements. Mat24 (and also, 
    to some extent, N_x0) is supported by class mat24.Mat24. 

    G_x0 has an extraspecial normal subgroup Q_x0 of shape 2**{1+24},
    see [Con85], [Seysen20]. Class Mat24Xi focusses on the operation of 
    xi and x**2 on Q_x0 by conjugation.

    This class provides a table-based implementation of the operation
    of xi and xi**2, which is based on [Seysen20], Lemma 9.5. The class 
    also provides some auxiliary functions used for that operation 
    defined in [Seysen20].
    
    The so-called short elements of Q_x0 (see [Con85], sect.7) play an 
    important role as basis vectors in the representation of the 
    monster in [Con85] and in this project. Class Mat24Xi also contains 
    functions for mapping short vectors in Q_x0 to basis vectors of 
    that representation and vice versa. The basis for our representa-
    tion of the monster group, is described in module mm_aux.c.

    A faster C implementation of this class is given in class
    mat24fast.Mat24Xi      

    References
    ----------
    see file refereces.txt
    """
    tab_g_gray = numpy.zeros(64, dtype = numpy.uint8)
    tab_g_cocode = numpy.zeros(64, dtype = numpy.uint8)
    #tab_g_col = numpy.zeros(64, dtype = numpy.uint8)

    tables = {
            "GenXi_module_doc" : globals()["__doc__"],
            "GenXi_g_gray_table" : tab_g_gray,
            "GenXi_g_cocode_table" : tab_g_cocode,
    }

    directives = {}

    for x in range(64):
        w2, c = w2_gamma(expand_gray(x))
        cx = compress_gray(c)
        w2x = w2 << 7 
        tab_g_gray[x] = w2x + cx 
        tab_g_cocode[cx] = w2x + x


    

    @classmethod
    def gen_xi_g_gray(cls, v):
        """Implement function gamma() in [Seysen20], section 3.3.

        Given a Golay code vector v  in 'gcode' representation, see
        module mat24.py, we return the element gamma(v) of the Golay 
        cocode, with gamma() as in [Seysen20], section 3.3. c is 
        returned in 'cocode' representation, see module mat24.py.
        """
        return expand_gray(cls.tab_g_gray[compress_gray(v)])
  
    @classmethod
    def gen_xi_w2_gray(cls, v):
        """Implement function w2() in [Seysen20], section 3.3.

        Given a Golay code vector v  in 'gcode' representation, see
        module mat24.py, we return the bit w2 = w2(c), with w2() as
        defined in [Seysen20], section 3.3.   
        """
        return cls.tab_g_gray[compress_gray(v)] >> 7

    @classmethod
    def gen_xi_g_cocode(cls, c):
        """Inverse of method xi_g_gray(v)

        Given a cocode vector c in in 'cocode' representation, see 
        module mat24.py, the function returns the unique gray Golay
        code vector v such that cls.gen_xi_w2_gray(v) is the gray 
        part of c. 
        """
        return expand_gray(cls.tab_g_cocode[compress_gray(c)])
  
    @classmethod
    def gen_xi_w2_cocode(cls, c):
        """Implement function w2() in [Seysen20], section 3.3.

        Given a cocode vector c in in 'cocode' representation, see 
        module mat24.py, the function returns the bit w2 = w2(c), 
        with w2() as defined in [Seysen20], section 3.3.   
        """
        return cls.tab_g_cocode[compress_gray(c)] >> 7

    @classmethod
    def gen_xi_op_xi(cls, x, exp):
        """Operation of  xi**exp  on the element x of the group Q_x0.

        The function returns the element

            xi**(-exp)  *  x  *  xi**exp

        of the group Q_x0. The element x of Q_x0 as must be given in
        Leech lattice encoding, as described in the header of this 
        module.

        The returned result is is coded in the same way. This 
        function implements the formula in [Seysen20], Lemma 9.5.
        """
        exp %= 3
        if (exp == 0): 
            return x
        scal = bw24((x >> 12) &  x & 0xc0f) & 1  
        x ^= scal << 24 # xor scalar product to sign

        tv =  cls.tab_g_gray[compress_gray(x >> 12)] 
        w2v, gv = tv >> 7, expand_gray(tv)
        tc =  cls.tab_g_cocode[compress_gray(x)] 
        w2c, gc = tc >> 7, expand_gray(tc)
        if (exp == 1):
            x &= ~0xc0f000  # kill gray code part
            x ^= w2c << 24  # xor w2(cocode) to sign
        else:
            x &= ~0xc0f     # kill gray cocode part
            x ^= w2v << 24  # xor w2(code) to sign
        x ^= gv         # xor g(code) to cocode 
        x ^= gc << 12   # xor g(cocode) to code
        return x




        
    @staticmethod
    def gen_xi_short_to_leech(x1):
        """Convert short vector to Leech lattice encoding.

        Both, Leech lattice and short vector encoding of a short vector 
        in Q_x are decribed in the header of this module. The function 
        returns the Leech lattice encoding of element x1 given in short
        vector encoding. 

        The function returns 0 for an illegal input x1. 
        """   
        box, sign, code = x1 >> 16, (x1 >> 15) & 1, x1 & 0x7fff
        octad = 0xffff
        if box == 1:
            if code < 1536:  # this is 2 * 24 * 32 
                gcode = code >= 768
                if gcode:
                    code -= 768
                gcode <<= 11
                i, j = code >> 5, code & 31
                cocode = Mat24.vect_to_cocode((1 << i) ^ (1 << j))
                if cocode == 0 or cocode & 0x800:
                    return 0
            elif code < 2496: # this is 2 * 24 * 32 + 15 * 64 
                octad = code - 1536
            else:
                return 0
        elif box == 2:
            if code >= 23040:  # this is  360 * 64     
                return 0   
            octad = code + 960  # this is 15 * 64
        elif box == 3:
            if code >= 24576:  # this is  384 * 64     
                return 0   
            octad = code + 24000  # this is (15 + 360) * 64
        elif box < 6:
            code += (box - 4) << 15
            cocode = Mat24.vect_to_cocode(1 << (x1 & 31))
            if cocode  == 0:
                return 0
            gcode = (code >> 5) & 0x7ff
            w =  Mat24.gcode_weight(gcode) ^ Mat24.scalar_prod(gcode, cocode) 
            gcode ^= (w & 1) << 11
        else:
            return 0  
        if octad < 48756:  # this is  759 * 64 
            cc = octad & 0x3f
            w = Mat24.bw24(cc) 
            cc = (cc << 1) ^ (w & 1)
            w += w & 1
            gcode = Mat24.octad_to_gcode(octad >> 6)
            gcodev = Mat24.gcode_to_vect(gcode)
            cocode = Mat24.vect_to_cocode(Mat24.spread_b24(cc, gcodev))
            gcode ^=  ((w >> 1) & 1) << 11
        cocode ^= Mat24.ploop_theta(gcode)
        return (sign << 24) | (gcode << 12) | cocode

    @staticmethod
    def gen_xi_leech_to_short(x1):   
        """Convert Leech lattice to short vector encoding.

        Both, Leech lattice and short vector encoding of a short vector 
        in Q_x are decribed in the header of this module. The function 
        returns the short vector encoding of element x1 given in Leech 
        lattice encoding. 

        The function returns 0 if the vector x1 is not short. 
        """   
        sign = (x1 >> 24) & 1
        x1 ^= Mat24.ploop_theta(x1 >> 12)
        gcodev = Mat24.gcode_to_vect(x1 >> 12) 
        cocodev = Mat24.cocode_syndrome(x1, 
            min(23, Mat24.lsbit24(gcodev))
        )
        w = Mat24.gcode_weight(x1 >> 12)
        if x1 & 0x800:
            if (Mat24.bw24(cocodev) > 1 or
                Mat24.scalar_prod(x1 >> 12, x1) !=  (w & 1)):
                return 0
            y = Mat24.lsbit24(cocodev)
            code = (x1 & 0x7ff000) >> 7 | y  
            box = 4 + (code >> 15)
            code &= 0x7fff
        else: 
            if w == 3:
                return 0
            elif w in [2,4]:
                if w == 4:
                    gcodev ^= 0xffffff
                    x1 ^= 0x800000
                w_bad = (Mat24.bw24(cocodev) ^ 2 ^ w) & 3
                if w_bad or (cocodev & gcodev) != cocodev:
                    return 0
                c =  Mat24.extract_b24(cocodev, gcodev)
                if (c & 0x80):
                     c ^=  0xff 
                y1 = Mat24.gcode_to_octad(x1 >> 12)
                code = (y1 << 6) | (c >> 1)
                if code >= 24000: # this is (15 + 360) * 64
                    code -= 24000
                    box = 3
                elif code >= 960: # this is 15 * 64
                    code -= 960
                    box = 2
                else:
                    code += 1536
                    box = 1
            else:
                y1 = Mat24.lsbit24(cocodev) 
                cocodev ^= 1 << y1  
                y2 = Mat24.lsbit24(cocodev) 
                if cocodev != (1 << y2) or y1 >= 24:
                    return 0
                code = 384 * (w & 2) + 32 * y2 + y1
                box = 1
        return  (box << 16) | (sign << 15) | code
  

    @classmethod
    def gen_xi_op_xi_short(cls, x, exp):
        """Operation of  xi**exp  on the element x of the group Q_x0.

        The function returns the element

            xi**(-exp)  *  x  *  xi**exp

        of the group Q_x0. The element x of Q_x0 as must be given in
        short vector encoding, as described in the header of this 
        module.

        The returned result is is coded in the same way. 
        """
        y = cls.gen_xi_short_to_leech(x)
        if (y == 0): return x
        y = cls.gen_xi_op_xi(y, exp)
        if (y == 0): return x
        y =  cls.gen_xi_leech_to_short(y)
        if (y == 0): return x
        return y

        
    @classmethod
    def make_table(cls, u_box, u_exp):
        assert 1 <= u_box <= 5 and 1 <= u_exp <= 2
        t_size = [0, 2496, 23040, 24576, 32768, 32768]
        a = numpy.zeros(32768, dtype = numpy.uint16) 
        length = t_size[u_box] 
        u_box <<= 16;
        for i  in range(length):
            a[i] = cls.gen_xi_op_xi_short(u_box + i, u_exp) & 0xffff;
        return a[:length]

    @staticmethod
    def invert_table(table, n_columns, len_result):
        assert len(table) & 31 == 0 and len_result & 31 == 0
        result = numpy.zeros(len_result, dtype = numpy.uint16) 
        for i, r in enumerate(table):
            if ((i & 31) < n_columns and (r & 0x7fff) < len_result):
                result[r & 0x7fff] = i | (r & 0x8000)
        return result


    @staticmethod
    def split_table(table, modulus):
        length = len(table)
        assert length & 31 == 0
        a = numpy.zeros(length >> 5, dtype = numpy.uint32) 
        for i in range(0, length, 32):
            sign = sum(((table[i+j] >> 15) & 1) << j for j in range(32))
            a[i >> 5] = sign
        table = (table & 0x7fff) % modulus
        return table, a





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import collections
import re
import warnings
from random import randint, shuffle
import pytest
import numpy as np

from mmgroup import MMSpace, MMGroup, MM0, MMV
from mmgroup.tests.spaces.leech_space import mul24_mmgroup


M = MM0
V = MMV(3)


def check_abs_equal(v, v_ref):
    if (v == v_ref).all():
        return 
    if (v == -v_ref).all():
        return 
    print("Leech lattice vector expected:")
    print(v_ref)
    print("Leech lattice vector obtained:")
    print(v)
    raise ValueError("Leech vectors are not equal or opposite")
        

def do_test_leech_variants(v, v_leech):
    index = MMSpace.tuple_to_index(v)
    v1 = MMSpace.index_to_short(index)
    check_abs_equal(v1, v_leech)
    tuple_= MMSpace.index_to_tuple(v)
    v2 = MMSpace.index_to_short(*tuple_)
    check_abs_equal(v2, v_leech)



def one_test_leech_op(v, g, verbose = 0):
    """``v`` a vector in ``V``, ``g`` an element of ``M``""" 
    #assert v in V
    #assert g in M
    v_leech =  MMSpace.index_to_short(v)
    do_test_leech_variants(v, v_leech)
    vg = v * g
    vg_leech_ref = MMSpace.index_to_short(vg)
    if verbose:
        print("\ntest ", g)
        print(v_leech)
        print(vg_leech_ref)
    vg_leech = mul24_mmgroup(v_leech, g)
    if verbose:
        print(vg_leech)
    check_abs_equal(vg_leech, vg_leech_ref)


def leech_op_testdata(n_tests = 100):
    for i in range(n_tests):
        for vtag in "BCTTTXXX":
            v = V(vtag,'r')
            l = [x for x in  "dpyxl"]
            shuffle(l)
            g = M([(x, 'r') for x in l])
            yield v, g
    



@pytest.mark.space
def test_leech_op_xy(verbose = 0):
    print("Check vector against Leech lattice operation ...")
    for v,g in leech_op_testdata():
        one_test_leech_op(v, g)
    print("passed")






